#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 28.04.2018 19:06
:Licence MIT
Part of grammpy

"""

from .GrammpyException import GrammpyException


class StartSymbolNotSetException(GrammpyException, ValueError):
    """
    Represent error when the start symbol of the grammar is not set and is required.
    """
    pass



#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 04.07.2017 19:45
:Licence MIT
Part of grammpy

"""
from typing import Any

from .support._RuleConnectable import _RuleConnectable


class Terminal(_RuleConnectable):
    '''
    Represent Terminal in formal grammar.
    '''

    def __init__(self, symbol):
        # type: (Any) -> Terminal
        '''
        Create new terminal.
        :param symbol: Symbol representing terminal.
        '''
        super().__init__()
        self.__symbol = symbol

    def __hash__(self):
        # type: () -> int
        '''
        Return hash based on symbol (symbol's hash).
        :return: Hash value of the terminal.
        '''
        return hash(self.__symbol)

    def __eq__(self, other):
        # type: (Any) -> bool
        '''
        Compare two terminals based on hash values.
        :param other: Terminal to compare.
        :return: True if objects are equal, false otherwise.
        '''
        return hash(self) == hash(other)

    def symbol(self):
        # type: () -> Any
        '''
        Get symbol for the terminal.
        :return: Symbol that represent current terminal.
        '''
        return self.__symbol

    @property
    def s(self):
        # type: () -> Any
        """
        Shortcut for symbol method.
        :return: Symbol that represent current terminal.
        """
        return self.symbol()



#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 17.08.207 13:29
:Licence MIT
Part of grammpy

"""
from copy import copy
from typing import TYPE_CHECKING

from ...exceptions import StartSymbolNotSetException

if TYPE_CHECKING:  # pragma: no cover
    from ... import Grammar


def remove_unreachable_symbols(grammar, inplace=False):
    # type: (Grammar, bool) -> Grammar
    """
    Remove unreachable symbols from the gramar
    :param grammar: Grammar where to symbols remove
    :param inplace: True if transformation should be performed in place. False by default.
    :return: Grammar without unreachable symbols.
    """
    # copy if required
    if inplace is False:
        grammar = copy(grammar)
    # check if start symbol is set
    if grammar.start is None:
        raise StartSymbolNotSetException()
    # create process sets
    reachable = {grammar.start}
    rules = grammar.rules.copy()
    # begin iterations
    while True:
        # create sets for current iteration
        active = reachable.copy()
        # loop the working rules
        for rule in rules.copy():
            # lf left part of rule already in reachable symbols
            if rule.fromSymbol in reachable:
                # set symbols on the right as reachable
                for symbol in rule.right:
                    active.add(symbol)
                # remove rule from the next iteration
                rules.remove(rule)
            # end of rules loop
        # if current and previous iterations are same, we are done
        if active == reachable:
            break
        # otherwise swap the sets
        reachable = active
    # remove the symbols
    nonterminals_to_remove = grammar.nonterminals.difference(reachable)
    terminals_to_remove = grammar.terminals.difference(reachable)
    grammar.nonterminals.remove(*nonterminals_to_remove)
    grammar.terminals.remove(*terminals_to_remove)
    # return grammar
    return grammar



#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 01.08.2017 07:33
:Licence MIT
Part of grammpy

"""
from typing import Type, Optional, Iterable, TYPE_CHECKING, Any

from .Nonterminal import Nonterminal
from .support._NonterminalSet import _NonterminalSet
from .support._RulesSet import _RulesSet
from .support._TerminalSet import _TerminalSet
from .support._constants import EPSILON
from ..exceptions import NonterminalDoesNotExistsException

if TYPE_CHECKING:  # pragma: no cover
    from . import Rule


class Grammar:
    """
    Provide interface for manipulating with the grammar
    """

    def __init__(self,
                 terminals=None,
                 nonterminals=None,
                 rules=None,
                 start_symbol=None):
        # type: (Optional[Iterable[Any]], Optional[Iterable[Type[Nonterminal]]], Optional[Iterable[Type[Rule]]], Optional[Type[Nonterminal]]) -> Grammar
        """
        Create instance of grammar.
        :param terminals: Sequence of terminals to add, empty sequence by default.
        :param nonterminals: Sequence of nonterminals to add, empty sequence by default.
        :param rules: Sequence of rules to add, empty sequence by default.
        :param start_symbol: Start symbol of the grammar, start symbol must be in nonterminals.
        None by default.
        """
        assign_map = dict({EPSILON: set()})
        self._terminals = _TerminalSet(self, assign_map, terminals)
        self._nonterminals = _NonterminalSet(self, assign_map, nonterminals)
        self._start_symbol = None
        self._rules = _RulesSet(self, assign_map, rules)
        self.start = start_symbol

    @property
    def terminals(self):
        # type: () -> _TerminalSet
        """
        Get set representing terminals.
        :return: Set representing terminals.
        """
        return self._terminals

    @property
    def nonterminals(self):
        # type: () -> _NonterminalSet
        """
        Get set representing nonterminals.
        :return: Set representing nonterminals.
        """
        return self._nonterminals

    @property
    def rules(self):
        # type: () -> _RulesSet
        """
        Get set representing rules.
        :return: Set representing rules.
        """
        return self._rules

    @property
    def start(self):
        # type: () -> Optional[Type[Nonterminal]]
        """
        Get start symbol of the grammar.
        :return: Start symbol of the grammar, or None if it isn't defined.
        """
        return self._start_symbol

    @start.setter
    def start(self, s):
        # type: (Optional[Type[Nonterminal]]) -> None
        """
        Set start symbol of the grammar.
        :param s: Start symbol to set.
        :raise NonterminalDoesNotExistsException: If the start symbol is not in nonterminals.
        """
        if s is not None and s not in self.nonterminals:
            raise NonterminalDoesNotExistsException(None, s, self)
        self._start_symbol = s

    @start.deleter
    def start(self):
        # type: () -> None
        """
        Unset start symbol and set it to None.
        """
        self.start = None

    def __copy__(self):
        # type: () -> Grammar
        """
        Perform shallow copy of the grammar.
        :return: Shallow copy of the grammar.
        """
        return Grammar(terminals=self.terminals,
                       nonterminals=self.nonterminals,
                       rules=self.rules,
                       start_symbol=self.start)



#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 20.08.2017 15:43
:Licence MIT
Part of grammpy

"""
from copy import copy
from queue import Queue
from typing import TYPE_CHECKING, Type, Dict

from .find_nonterminals_rewritable_to_epsilon import find_nonterminals_rewritable_to_epsilon
from ... import Rule, EPSILON

if TYPE_CHECKING:  # pragma: no cover
    from ... import Grammar, Nonterminal


class EpsilonRemovedRule(Rule):
    """
    Replace rule, when one symbol of the original rule is rewritable to epsilon.
    The new Rule contains original rule in `from_rule` field,
    index of the nonterminal rewritable to epsilon at `replace_index` field and
    series of rules ending the epsilon in `backtrack` field.
    """
    from_rule = None  # type: Type[Rule]
    replace_index = None  # type: int
    backtrack = None  # type: Dict[Type[Nonterminal], Type[Rule]]


def _create_rule(rule, index, backtrack):
    # type: (Rule, int, Dict[Type[Nonterminal], Type[Rule]]) -> Type[EpsilonRemovedRule]
    """
    Create EpsilonRemovedRule. This rule will skip symbol at the `index`.
    :param rule: Original rule.
    :param index: Index of symbol that is rewritable to epsilon.
    :param backtrack: Dictionary where key is nonterminal and value is rule which is next to generate epsilon.
    :return: EpsilonRemovedRule class without symbol rewritable to epsilon.
    """
    # remove old rules from the dictionary
    old_dict = rule.__dict__.copy()
    if 'rules' in old_dict: del old_dict['rules']
    if 'rule' in old_dict: del old_dict['rule']
    if 'left' in old_dict: del old_dict['left']
    if 'right' in old_dict: del old_dict['right']
    if 'fromSymbol' in old_dict: del old_dict['fromSymbol']
    if 'toSymbol' in old_dict: del old_dict['toSymbol']
    # create type
    created = type('NoEps[' + rule.__name__ + ']',
                   (EpsilonRemovedRule,),
                   old_dict)  # type: Type[EpsilonRemovedRule]
    # add from_rule and index
    created.from_rule = rule
    created.replace_index = index
    created.backtrack = backtrack
    # attach rule
    created.fromSymbol = rule.fromSymbol
    created.right = [rule.right[i] for i in range(len(rule.right)) if i != index]
    # ff the right side is empty
    if len(created.right) == 0:
        created.right = [EPSILON]
    return created


def remove_rules_with_epsilon(grammar, inplace=False):
    # type: (Grammar, bool) -> Grammar
    """
    Remove epsilon rules.
    :param grammar: Grammar where rules remove
    :param inplace: True if transformation should be performed in place, false otherwise.
    False by default.
    :return: Grammar without epsilon rules.
    """
    # copy if required
    if inplace is False:
        grammar = copy(grammar)
    # find nonterminals rewritable to epsilon
    rewritable = find_nonterminals_rewritable_to_epsilon(grammar)  # type: Dict[Type[Nonterminal], Type[Rule]]
    # create queue from rules to iterate over
    rules = Queue()
    for r in grammar.rules:
        rules.put(r)
    # iterate thought rules
    while not rules.empty():
        rule = rules.get()
        right = rule.right
        # if the rule rewrite to epsilon we can safely delete it
        if right == [EPSILON]:
            # unless it rewrites from the start symbol
            if rule.fromSymbol != grammar.start:
                grammar.rules.discard(rule)
            # continue IS executed, but due optimization line is marked as missed.
            continue  # pragma: no cover
        # iterate over the right side
        for rule_index in range(len(right)):
            symbol = right[rule_index]
            # if symbol is rewritable, generate new rule without that symbol
            if symbol in rewritable:
                new_rule = _create_rule(rule, rule_index, rewritable)
                grammar.rules.add(new_rule)
                rules.put(new_rule)  # in case there are more rewritable symbols
    return grammar



#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 31.01.2019 13:48
:Licence MIT
Part of grammpy

"""
import weakref
from typing import TYPE_CHECKING, Optional

from ...exceptions import TreeDeletedException

if TYPE_CHECKING:  # pragma: no cover
    from .. import Rule


class _RuleConnectable:
    """
    Represent entity that can be connected by Rule in the AST.
    """

    def __init__(self):
        """
        Creates new instance.
        Properties from_rule and to_rule are None.
        """
        self._from_rule = None
        self._to_rule = None

    @property
    def from_rule(self):
        # type: (_RuleConnectable) -> Optional[Rule]
        """
        Rule that the symbol is rewrote from.
        :return: Instance of Rule that the symbol is rewrited from.
        :raise TreeDeletedException: If the tree was already deleted (pointer to the painter get lost).
        This can occur because weak reference to the parent is used.
        """
        if self._from_rule is None:
            return None
        if self._from_rule() is None:
            raise TreeDeletedException()
        return self._from_rule()

    @property
    def to_rule(self):
        # type: (_RuleConnectable) -> Optional[Rule]
        """
        Rule that the symbol is rewrote to.
        :return: Instance of Rule that the symbols is rewrited to.
        """
        return self._to_rule

    def _set_from_rule(self, r):
        # type: (_RuleConnectable, Rule) -> None
        """
        Set the rule that the symbol is rewrote from.
        :param r: Instance of Rule that the symbol should rewrite from.
        """
        self._from_rule = weakref.ref(r)

    def _set_to_rule(self, r):
        # type: (_RuleConnectable, Rule) -> None
        """
        Set the rule that the symbol is rewrote to.
        :param r: Instance of Rule that the symbol should rewrite to.
        """
        self._to_rule = r



#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 01.08.2017 08:46
:Licence MIT
Part of grammpy

"""

from .CannotConvertException import CannotConvertException


class NotASingleSymbolException(CannotConvertException):
    """
    More symbols defined at the place where one symbol is expected
    """
    def __init__(self, symbols):
        self.symbols = symbols



#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 08.07.2017 14:03
:Licence MIT
Part of grammpy

"""

from .CannotConvertException import CannotConvertException
from .CantCreateSingleRuleException import CantCreateSingleRuleException
from .GrammpyException import GrammpyException
from .MultipleDefinitionException import MultipleDefinitionException
from .NonterminalDoesNotExistsException import NonterminalDoesNotExistsException
from .NotASingleSymbolException import NotASingleSymbolException
from .NotNonterminalException import NotNonterminalException
from .NotParsedException import NotParsedException
from .NotRuleException import NotRuleException
from .RuleException import RuleException
from .RuleNotDefinedException import RuleNotDefinedException
from .RuleSyntaxException import RuleSyntaxException
from .StartSymbolNotSetException import StartSymbolNotSetException
from .TerminalDoesNotExistsException import TerminalDoesNotExistsException
from .TreeDeletedException import TreeDeletedException
from .UselessEpsilonException import UselessEpsilonException



#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 03.09.2017 10:46
:Licence MIT
Part of grammpy

"""
from collections import deque
from typing import TYPE_CHECKING

from .transform_to_chomsky_normal_form import *
from ..Manipulations import Manipulations
from ..Traversing import Traversing

if TYPE_CHECKING:  # pragma: no cover
    from ... import Nonterminal


def transform_from_chomsky_normal_form(root):
    # type: (Nonterminal) -> Nonterminal
    """
    Transform the tree created by grammar in the Chomsky Normal Form to original rules.
    :param root: Root of parsed tree.
    :return: Modified tree.
    """
    # Transforms leaves
    items = Traversing.post_order(root)
    items = filter(lambda x: isinstance(x, (ChomskyTermRule, ChomskyTerminalReplaceRule)), items)
    de = deque(items)
    while de:
        rule = de.popleft()
        if isinstance(rule, ChomskyTermRule):
            upper_nonterm = rule.from_symbols[0]  # type: Nonterminal
            term = rule.to_symbols[0]
            Manipulations.replaceNode(upper_nonterm, term)
        elif isinstance(rule, ChomskyTerminalReplaceRule):
            created_rule = rule.from_rule()  # type: Rule
            Manipulations.replaceRule(rule, created_rule)
            de.append(created_rule)

    # Transform inner nodes
    items = Traversing.post_order(root)
    items = filter(lambda x: isinstance(x, ChomskySplitRule), items)
    de = deque(items)
    while de:
        rule = de.popleft()
        if isinstance(rule, ChomskySplitRule):
            created_rule = rule.from_rule()  # type: Rule
            # parent nonterminals
            for p in rule.from_symbols:  # type: Nonterminal
                p._set_to_rule(created_rule)
                created_rule._from_symbols.append(p)
            # left child
            left_child = rule.to_symbols[0]  # type: Nonterminal
            left_child._set_from_rule(created_rule)
            created_rule._to_symbols.append(left_child)
            # right childs
            for ch in rule.to_symbols[1].to_rule.to_symbols:  # type: Nonterminal
                ch._set_from_rule(created_rule)
                created_rule.to_symbols.append(ch)
            # add back if the rules is ChomskySplitRule again
            de.appendleft(created_rule)
    return root



#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 01.08.2017 08:54
:Licence MIT
Part of grammpy

"""

from .GrammpyException import GrammpyException

class CannotConvertException(GrammpyException, ValueError):
    """
    Library cant convert objects
    """
    pass


#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 22.08.2017 11:39
:Licence MIT
Part of grammpy

"""

from typing import List, Dict, Type, TYPE_CHECKING

from ._support import _is_unit
from ... import Grammar, Nonterminal

if TYPE_CHECKING:  # pragma: no cover
    from ... import Rule

    MATRIX_OF_UNIT_RULES = List[List[List[Type[Rule]]]]


class UnitSymbolReachability:
    """
    Class containing info about reachability of unit rules.
    """

    def __init__(self, field, translation):
        # type: (MATRIX_OF_UNIT_RULES, Dict[Type[Nonterminal], int]) -> None
        """
        Create isntance of UnitSymbolRechablingResults
        :param field: Result of Floyd-Warshall algorithm
        :param translation: Dictionary where keys is nonterminal and value is position in the field.
        """
        # matrix of lists
        self.f = field  # type: MATRIX_OF_UNIT_RULES
        self.t = translation  # type: Dict[Type[Nonterminal], int]

    def reach(self, from_symbol, to_symbol):
        # type: (Type[Nonterminal], Type[Nonterminal]) -> bool
        """
        Check if exists sequence of unit rules between two symbols.
        :param from_symbol: From which symbol find the sequence.
        :param to_symbol:  To which symbol find the sequence.
        :return: True if exists sequence of unit rules, false otherwise.
        """
        return len(self.path_rules(from_symbol, to_symbol)) > 0

    def reachables(self, from_symbol):
        # type: (Type[Nonterminal]) -> List[Type[Nonterminal]]
        """
        Get list of nonterminals, what are rewritable from nonterminal passed as parameter
        using only unit rules.
        :param from_symbol: For which symbols to search.
        :return: List of nonterminals.
        """
        if from_symbol not in self.t:
            return []
        reachable = []
        index = self.t[from_symbol]
        for n, i in self.t.items():
            if len(self.f[index][i] or []) > 0:
                reachable.append(n)
        return reachable

    def path_rules(self, from_symbol, to_symbol):
        # type: (Type[Nonterminal], Type[Nonterminal]) -> List[Type[Rule]]
        """
        Get sequence of unit rules between first and second parameter.
        :param from_symbol: From which symbol.
        :param to_symbol: To which symbol.
        :return: Sequence of unit rules. Empty sequence mean there is no way between them.
        """
        if from_symbol not in self.t or to_symbol not in self.t:
            return []
        return self.f[self.t[from_symbol]][self.t[to_symbol]] or []


def find_nonterminals_reachable_by_unit_rules(grammar):
    # type: (Grammar) -> UnitSymbolReachability
    """
    Get nonterminal for which exist unit rule
    :param grammar: Grammar where to search
    :return: Instance of UnitSymbolReachability.
    """
    # get nonterminals
    nonterminals = list(grammar.nonterminals)  # type: List[Type[Nonterminal]]
    count_of_nonterms = len(nonterminals)
    # create indexes for nonterminals
    nonterm_to_index = dict()  # type: Dict[Type[Nonterminal], int]
    for i in range(count_of_nonterms):
        nonterm_to_index[nonterminals[i]] = i
    # prepare matrix
    field = [[None for _ in nonterminals] for _ in nonterminals]  # type: MATRIX_OF_UNIT_RULES
    # fill existing unit rules
    for rule in grammar.rules:
        if _is_unit(rule):
            field[nonterm_to_index[rule.fromSymbol]][nonterm_to_index[rule.toSymbol]] = [rule]
    # run Floyd Warshall
    f = field
    for k in range(count_of_nonterms):
        for i in range(count_of_nonterms):
            for j in range(count_of_nonterms):
                if f[i][k] is not None and f[k][j] is not None:
                    if f[i][j] is None or len(f[i][j]) > len(f[i][k]) + len(f[k][j]):
                        f[i][j] = f[i][k] + f[k][j]
    # return results
    return UnitSymbolReachability(f, nonterm_to_index)



#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 02.08.2017 17:25
:Licence MIT
Part of grammpy

"""

from .RuleException import RuleException

class RuleSyntaxException(RuleException):
    """
    Syntax of the rule or rules is invalid
    """
    def __init__(self, rule, message, additional = None):
        super().__init__(rule)
        self.message = message
        self.additional = additional




#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 20.08.2017 15:42
:Licence MIT
Part of grammpy

"""
from typing import TYPE_CHECKING, Dict, Type, Optional

from ... import EPSILON

if TYPE_CHECKING:  # pragma: no cover
    from ... import Grammar, Nonterminal, Rule


def find_nonterminals_rewritable_to_epsilon(grammar):
    # type: (Grammar) -> Dict[Type[Nonterminal], Type[Rule]]
    """
    Get nonterminals rewritable to epsilon.
    :param grammar: Grammar where to search.
    :return: Dictionary, where key is nonterminal rewritable to epsilon and value is rule that is responsible for it.
    The rule doesn't need to rewrite to epsilon directly,
    but the whole right side can be rewritable to epsilon using different rules.
    """
    # start with empty dictionary (contains only epsilon)
    rewritable = dict()  # type: Dict[Type[Nonterminal], Optional[Type[Rule]]]
    rewritable[EPSILON] = None
    # iterate until the dictionary change
    while True:
        working = rewritable.copy()
        for rule in grammar.rules:
            # no need to process rules we already know rewrite to epsilon
            if rule.fromSymbol in working:
                continue
            # check if the whole right side rewrite to epsilon
            right_side_rewrite = True
            for symbol in rule.right:
                if symbol not in rewritable:
                    right_side_rewrite = False
            # the whole right side can be reduce to epsilon, add left side to dictionary
            if right_side_rewrite:
                working[rule.fromSymbol] = rule
        # Working set didn't change, we are done
        if working == rewritable:
            break
        # Otherwise swap the sets and iterate
        rewritable = working
    # delete epsilon from the dictionary
    del rewritable[EPSILON]
    return rewritable



#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 22.08.2017 11:37
:Licence MIT
Part of grammpy

"""
from copy import copy
from typing import TYPE_CHECKING, List, Type

from ._support import _is_unit
from .find_symbols_reachable_by_unit_rules import find_nonterminals_reachable_by_unit_rules
from ... import Rule

if TYPE_CHECKING:  # pragma: no cover
    from ... import Grammar


class ReducedUnitRule(Rule):
    """
    Represent rule that replace sequence of unit rules (with the last rule that generates something)
    with single rule.
    The sequence of applied rules is in the `by_rules` property.
    The last, generating rule is in the `end_rule` property.
    """
    by_rules = []  # type: List[Type[Rule]]
    end_rule = None  # type: Type[Rule]


def _create_rule(path, rule):
    # type: (List[Type[Rule]], Type[Rule]) -> Type[ReducedUnitRule]
    """
    Create ReducedUnitRule based on sequence of unit rules and end, generating rule.
    :param path: Sequence of unit rules.
    :param rule: Rule that is attached after sequence of unit rules.
    :return: ReducedUnitRule class.
    """
    created = type('Reduced[' + rule.__name__ + ']',
                   (ReducedUnitRule,),
                   ReducedUnitRule.__dict__.copy())  # type: Type[ReducedUnitRule]
    created.rule = ([path[0].fromSymbol], rule.right)
    created.end_rule = rule
    created.by_rules = path
    return created


def remove_unit_rules(grammar, inplace=False):
    # type: (Grammar, bool) -> Grammar
    """
    Remove unit rules from the grammar.
    :param grammar: Grammar where remove the rules.
    :param inplace: True if transformation should be performed in place. False by default.
    :return: Grammar without unit rules.
    """
    # copy if needed
    if inplace is False:
        grammar = copy(grammar)
    # get connections
    res = find_nonterminals_reachable_by_unit_rules(grammar)
    # iterate through rules
    for rule in grammar.rules.copy():
        # delete unit rules
        if _is_unit(rule):
            grammar.rules.remove(rule)
            continue
        for nonterm in grammar.nonterminals:
            # find all nonterminals that can rewrite to current rule
            path = res.path_rules(nonterm, rule.fromSymbol)
            # get rid of cyclic paths
            if len(path) > 0 and path[0].fromSymbol != path[-1].toSymbol:
                created = _create_rule(path, rule)
                grammar.rules.add(created)
    return grammar



#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 02.08.2017 17:25
:Licence MIT
Part of grammpy

"""

from .RuleSyntaxException import RuleSyntaxException


class MultipleDefinitionException(RuleSyntaxException):
    """
    The rule is defined multiple times.
    """

    def __init__(self, rule, message, additional=None):
        super().__init__(rule, message, additional)



#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 02.08.2017 17:24
:Licence MIT
Part of grammpy

"""

from .GrammpyException import GrammpyException


class RuleException(GrammpyException, ValueError):
    """
    Error with the rule or rules
    """
    def __init__(self, rule):
        super().__init__()
        self.rule = rule



#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 31.01.2019 13:46
:Licence MIT
Part of grammpy

"""

from inspect import isclass
from typing import Any, List, TYPE_CHECKING

from ._constants import EPS
from ..Nonterminal import Nonterminal
from ...exceptions import *

if TYPE_CHECKING:  # pragma: no cover
    from .. import Grammar


class _MetaRule(type):
    """
    Metaclass for rule.
    Contains logic, how to compute toSymbol, fromSymbol, left, right, rule, rules properties.
    Contains logic about rule validation.
    """

    def __init__(cls, name, bases, dct):
        cls._traverse = False
        super(_MetaRule, cls).__init__(name, bases, dct)

    def __hash__(cls):
        # type: (_MetaRule) -> int
        """
        Get hash for Rule class.
        :return: Hash of the Rule class.
        """
        try:
            transformed = tuple((tuple(s for s in rule[0]), tuple(s for s in rule[1])) for rule in cls.rules)
            return hash(transformed)
        except (TypeError, RuleNotDefinedException):
            return super().__hash__()

    def __eq__(cls, other):
        # type: (_MetaRule, object) -> bool
        """
        Compare two Rule classes.
        :param other: Another Rule class.
        :return: True if both classes contains same rules, false otherwise.
        """
        return isclass(other) and hash(cls) == hash(other)

    @staticmethod
    def _get_toSymbol(cls):
        # type: (_MetaRule) -> object
        """
        Get symbol from which the rule is rewrote.
        :param cls: Rule for which return the symbol.
        :return: Symbol from which the rule is rewrote.
        :raise RuleNotDefinedException: If the rule is not defined.
        :raise CantCreateSingleRuleException: If the rule consists of more rules.
        :raise NotASingleSymbolException: If number of symbols on the left is more.
        """
        if cls._traverse:
            raise RuleNotDefinedException(cls)
        if len(cls.rules) > 1:
            raise CantCreateSingleRuleException(cls)
        right = cls.rules[0][1]
        if len(right) > 1:
            raise NotASingleSymbolException(right)
        return right[0]

    @staticmethod
    def _get_fromSymbol(cls):
        # type: (_MetaRule) -> object
        """
        Get symbol to which the rule is rewrote.
        :param cls: Rule for which return the symbol.
        :return: Symbol to which the rule is rewrote.
        :raise RuleNotDefinedException: If the rule is not defined.
        :raise CantCreateSingleRuleException: If the rule consists of more rules.
        :raise NotASingleSymbolException: If number of symbols on the right is more.
        """
        if cls._traverse:
            raise RuleNotDefinedException(cls)
        if len(cls.rules) > 1:
            raise CantCreateSingleRuleException(cls)
        left = cls.rules[0][0]
        if len(left) > 1:
            raise NotASingleSymbolException(left)
        return left[0]

    @staticmethod
    def _get_right(cls):
        # type: (_MetaRule) -> List[object]
        """
        Get right part of the rule.
        :param cls: Rule for which return the right side.
        :return: Symbols on the right side of the array.
        :raise RuleNotDefinedException: If the rule is not defined.
        :raise CantCreateSingleRuleException: If the rule consists of more rules.
        :raise NotASingleSymbolException: If number of symbols on the left is more.
        """
        if cls._traverse:
            return [cls.toSymbol]
        if len(cls.rules) > 1:
            raise CantCreateSingleRuleException(cls)
        return cls.rules[0][1]

    @staticmethod
    def _get_left(cls):
        # type: (_MetaRule) -> List[object]
        """
        Get left part of the rule.
        :param cls: Rule for which return the left side.
        :return: Symbols on the left side of the array.
        :raise RuleNotDefinedException: If the rule is not defined.
        :raise CantCreateSingleRuleException: If the rule consists of more rules.
        :raise NotASingleSymbolException: If number of symbols on the left is more.
        """
        if cls._traverse:
            return [cls.fromSymbol]
        if len(cls.rules) > 1:
            raise CantCreateSingleRuleException(cls)
        return cls.rules[0][0]

    @staticmethod
    def _get_rule(cls):
        # type: (_MetaRule) -> (List[object], List[object])
        """
        Get rule on the Rule class.
        :param cls: Rule for which return the rule.
        :return: Rule inside the Rule class.
        :raise RuleNotDefinedException: If the rule is not defined.
        :raise CantCreateSingleRuleException: If the rule consists of more rules.
        :raise NotASingleSymbolException: If number of symbols on the left is more.
        """
        if cls._traverse:
            return (cls.left, cls.right)
        if len(cls.rules) > 1:
            raise CantCreateSingleRuleException(cls)
        return cls.rules[0]

    @staticmethod
    def _get_rules(cls):
        # type: (_MetaRule) -> List[(List[object], List[object])]
        """
        Get rules on the Rule class.
        :param cls: Rule for which return the rules.
        :return: Rules inside the Rule class.
        :raise RuleNotDefinedException: If the rule is not defined.
        :raise CantCreateSingleRuleException: If the rule consists of more rules.
        :raise NotASingleSymbolException: If number of symbols on the left is more.
        """
        cls._traverse = True
        r = cls.rule
        cls._traverse = False
        return [r]

    def __getattr__(cls, name):
        # type: (_MetaRule, str) -> Any
        if name in {'toSymbol',
                    'fromSymbol',
                    'left',
                    'right',
                    'rule',
                    'rules'}:
            return getattr(_MetaRule, '_get_' + name)(cls)
        raise AttributeError(cls.__name__ + ' doesn\'t have attribute ' + name)

    @property
    def count(cls):
        # type: (_MetaRule) -> int
        """
        Get count of rules defined in the class.
        :return: Count of rules.
        :raise RuleNotDefinedException: If the rule is not defined.
        :raise CantCreateSingleRuleException: If the rule consists of more rules.
        :raise NotASingleSymbolException: If number of symbols on the left is more.
        """
        return len(cls.rules)

    @property
    def rules_count(cls):
        # type: (_MetaRule) -> int
        """
        Get count of rules defined in the class
        :return: Count of rules
        :raise RuleNotDefinedException: If the rule is not defined.
        :raise CantCreateSingleRuleException: If the rule consists of more rules.
        :raise NotASingleSymbolException: If number of symbols on the left is more.
        """
        return cls.count

    def _controlSide(cls, side, grammar):
        # type: (_MetaRule, List[object], Grammar) -> None
        """
        Validate one side of the rule.
        :param side: Iterable side of the rule.
        :param grammar: Grammar on which to validate.
        :raise RuleSyntaxException: If invalid syntax is use.
        :raise UselessEpsilonException: If useless epsilon is used.
        :raise TerminalDoesNotExistsException: If terminal does not exists in the grammar.
        :raise NonterminalDoesNotExistsException: If nonterminal does not exists in the grammar.
        """
        if not isinstance(side, list):
            raise RuleSyntaxException(cls, 'One side of rule is not enclose by list', side)
        if len(side) == 0:
            raise RuleSyntaxException(cls, 'One side of rule is not define', side)
        if EPS in side and len(side) > 1:
            raise UselessEpsilonException(cls)
        for symb in side:
            if isclass(symb) and issubclass(symb, Nonterminal):
                if symb not in grammar.nonterminals:
                    raise NonterminalDoesNotExistsException(cls, symb, grammar)
            elif symb is EPS:
                continue
            elif symb not in grammar.terminals:
                raise TerminalDoesNotExistsException(cls, symb, grammar)

    def validate(cls, grammar):
        # type: (_MetaRule, Grammar) -> None
        """
        Perform rules validation of the class.
        :param grammar: Grammar on which to validate.
        :raise RuleSyntaxException: If invalid syntax is used.
        :raise UselessEpsilonException: If epsilon used in rules in useless.
        :raise TerminalDoesNotExistsException: If terminal does not exists in the grammar.
        :raise NonterminalDoesNotExistsException: If nonterminal does not exists in the grammar.
        """
        # check if the rule is not defined multiple times
        defined = set(dir(cls))
        if 'rules' in defined and len(defined & {'rule', 'left', 'right', 'toSymbol', 'fromSymbol'}) > 0 or \
                'rule' in defined and len(defined & {'left', 'right', 'toSymbol', 'fromSymbol'}) > 0 or \
                'left' in defined and 'fromSymbol' in defined or \
                'right' in defined and 'toSymbol' in defined:
            raise MultipleDefinitionException(cls, 'Rule is defined multiple times')
        # check if the rule is defined properly
        all = cls.rules
        if not isinstance(all, list):
            raise RuleSyntaxException(cls, 'Rules property is not enclose in list')
        for rule in all:
            if not isinstance(rule, tuple):
                raise RuleSyntaxException(cls, 'One of the rules is not enclose in tuple', rule)
            if len(rule) != 2:
                raise RuleSyntaxException(cls, 'One of the rules does not have define left and right part', rule)
            left = rule[0]
            right = rule[1]
            cls._controlSide(left, grammar)
            cls._controlSide(right, grammar)
            if left == [EPS] and right == [EPS]:
                raise UselessEpsilonException(cls)

    def is_valid(cls, grammar):
        # type: (_MetaRule, Grammar) -> bool
        """
        Check if the rule is correctly defined.
        :param grammar: Grammar on which to validate.
        :return: True if have class correct rules, false otherwise.
        """
        try:
            cls.validate(grammar)
            return True
        except RuleException:
            return False



#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 23.12.2017 16:05
:Licence MIT
Part of grammpy

"""
from typing import TYPE_CHECKING

from ..Manipulations import Manipulations
from ..Traversing import Traversing
from ...representation.support.SplitRule import SplitRule

if TYPE_CHECKING:  # pragma: no cover
    from ... import Nonterminal


def splitted_rules(root):
    # type: (Nonterminal) -> Nonterminal
    """
    Replace SplittedRules in the parsed tree with the original one.
    This method is mandatory if you insert Rule class with multiple rules into the grammar.
    :param root: Root of the parsed tree.
    :return: Modified tree.
    """
    items = Traversing.post_order(root)
    items = filter(lambda x: isinstance(x, SplitRule), items)
    for i in items:
        # create the original rule
        newRule = i.from_rule()
        # replace it with the node in the tree
        Manipulations.replace(i, newRule)
    return root



#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 31.08.2017 11:10
:Licence MIT
Part of grammpy

"""

from .GrammpyException import GrammpyException


class TreeDeletedException(GrammpyException):
    """
    Exception that parent of the AST was already deleted
    """
    pass



#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 08.07.2017 14:03
:Licence MIT
Part of grammpy

"""
from typing import Any


class GrammpyException(Exception):
    """
    Base class for all exceptions
    """
    def __init__(self, *args: Any) -> None:
        super().__init__(*args)


#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 31.08.2017 21:56
:Licence MIT
Part of grammpy

"""

from typing import TYPE_CHECKING, Type

if TYPE_CHECKING:  # pragma: no cover
    from ... import Rule, Nonterminal


class PlaceItem:
    """
    Represent item in the CYK structure.
    """
    def __init__(self, r, t1, t2):
        # type: (Type[Rule], Type[Rule], Type[Rule]) -> None
        """
        Create instance of PlaceItem.
        :param r: Rule to store.
        :param t1: Rule to match first symbol on the right side.
        :param t2: Rule to match second symbol on the right side.
        """
        self.to_rule = [t1, t2]
        self.rule = r

    def __hash__(self):
        # type: () -> int
        """
        Return hash of stored rule.
        :return: Hash of stored rule.
        """
        return hash(self.rule)

    @property
    def fromSymbol(self):
        # type: () -> Type[Nonterminal]
        """
        Get symbol on the left side of the rule.
        :return: Left side of the rule.
        """
        return self.rule.fromSymbol



#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 31.08.2017 18:47
:Licence MIT
Part of grammpy

"""
from collections import namedtuple
from typing import Any, Type, TYPE_CHECKING, List, Dict, Set

if TYPE_CHECKING:  # pragma: no cover
    from ... import Rule
    from .PlaceItem import PlaceItem


class Point(namedtuple('Point', ['x', 'y'])):
    """
    Represent point in 2D space.
    """
    pass


class Field:
    """
    Represent structure on which CYK operates

        X (first) axis
        __________
      y | A B C D
      | | E F G
      V | H I
        | J

    f[2][1] = I
    """

    def __init__(self, size):
        # type: (int) -> None
        """
        Create structure with specific size.
        :param size: Size of the structure.
        """
        self._field = [[[] for _ in range(k, 0, -1)] for k in range(size, 0, -1)]

    def fill(self, term_dict, terms):
        # type: (Dict[int, Set[Type[Rule]]], Any) -> None
        """
        Fill first row of the structure witch nonterminal directly rewritable to terminal.
        :param term_dict: Dictionary of rules directly rewritable to terminal.
        Key is hash of terminal, value is set of rules with key terminal at the right side.
        :param terms: Input sequence of terminal.
        """
        for i in range(len(terms)):
            t = terms[i]
            self._field[0][i] += term_dict[hash(t)]

    def rules(self, x, y):
        # type: (int, int) -> List[Type[Rule]]
        """
        Get rules at specific position in the structure.
        :param x: X coordinate
        :param y: Y coordinate
        :return: List of rules
        """
        return [r for r in self._field[y][x]]

    def positions(self, x, y):
        # type: (int, int) -> List[(Point, Point)]
        """
        Get all positions, that can be combined to get word parsed at specified position.
        :param x: X coordinate.
        :param y: Y coordinate.
        :return: List of tuples with two Point instances.
        """
        return [(Point(x, v), Point(x + 1 + v, y - 1 - v)) for v in range(y)]

    def put(self, x, y, rules):
        # type: (int, int, List[PlaceItem]) -> None
        """
        Set possible rules at specific position.
        :param x: X coordinate.
        :param y: Y coordinate.
        :param rules: Value to set.
        """
        self._field[y][x] = rules



#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 03.09.2017 10:47
:Licence MIT
Part of grammpy

"""
from typing import TYPE_CHECKING, Type, Dict, Any, Union

from .remove_rules_with_epsilon import EpsilonRemovedRule
from ..Traversing import Traversing
from ... import EPSILON, Terminal

if TYPE_CHECKING:  # pragma: no cover
    from ... import Nonterminal, Rule


def _restore_tree_for(root, translate):
    # type: (Any, Dict[Type[Nonterminal], Type[Rule]]) -> Union[Nonterminal, Terminal]
    """
    Create part of AST that generate epsilon.
    :param root: Symbol in the original rule that results in epsilon.
    Can be Nonterminal or epsilon itself.
    :param translate: Dictionary where key is nonterminal and value is rule which is next to generate epsilon.
    :return: Nonterminal instance with part of AST generating epsilon.
    """
    # the symbol is epsilon directly, just return Terminal.
    if root is EPSILON:
        return Terminal(EPSILON)
    # create nonterminal
    created_nonterm = root()  # type: Nonterminal
    created_rule = translate[root]()  # type: Rule
    created_nonterm._set_to_rule(created_rule)
    created_rule._from_symbols.append(created_nonterm)
    # all symbols from the right are rewritable to epsilon, so we need to restore them as well
    for ch in created_rule.right:
        p = _restore_tree_for(ch, translate)  # type: Nonterminal
        p._set_from_rule(created_rule)
        created_rule._to_symbols.append(p)
    return created_nonterm


def epsilon_rules_restore(root):
    # type: (Nonterminal) -> Nonterminal
    """
    Transform parsed tree to contain epsilon rules originally removed from the grammar.
    :param root: Root of the parsed tree.
    :return: Modified tree including epsilon rules.
    """
    items = Traversing.post_order(root)
    items = filter(lambda x: isinstance(x, EpsilonRemovedRule), items)
    for rule in items:
        # create original rule
        created_rule = rule.from_rule()  # type: Rule
        # attach parrents parents
        for s in rule.from_symbols:  # type: Nonterminal
            s._set_to_rule(created_rule)
            created_rule._from_symbols.append(s)
        # attach children up to replace index (that will contain epsilon)
        for i in range(rule.replace_index):
            ch = rule.to_symbols[i]  # type: Nonterminal
            ch._set_from_rule(created_rule)
            created_rule._to_symbols.append(ch)
        # add symbols originally rewrote to epsilon
        symb = _restore_tree_for(created_rule.right[rule.replace_index], rule.backtrack)  # type: Nonterminal
        created_rule._to_symbols.append(symb)
        symb._set_from_rule(created_rule)
        # attach rest of children
        for i in range(rule.replace_index, len(rule.to_symbols)):
            ch = rule.to_symbols[i]  # type: Nonterminal
            ch._set_from_rule(created_rule)
            created_rule._to_symbols.append(ch)
    return root



#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 10.03.2019 13:20
:Licence MIT
Part of grammpy

"""
from typing import Iterable, Any, AbstractSet


class _BaseSet(set):
    def update(self, *s):
        # type: (Iterable[Any]) -> _BaseSet
        """
        Add elements from all the parameters.
        :param s: Iterable objects with elements to add.
        :return: Current instance with inserted elements.
        """
        for iterable in s:
            self.add(*iterable)
        return self

    def __ior__(self, s):
        # type: (AbstractSet[Any]) -> _BaseSet
        """
        Add elements from all the parameters.
        :param s: Iterable objects with elements to add.
        :return: Current instance with inserted elements.
        """
        return self.update(s)

    def symmetric_difference_update(self, other):
        # type: (Iterable[Any]) -> _BaseSet
        """
        Update the TerminalSet.
        Keep elements from self and other, but discard elements that are in both.
        :param other: Iterable object with elements to compare with.
        :return: Current instance with updated state.
        """
        intersect = self.intersection(other)
        self.remove(*intersect)
        for elem in set(other).difference(intersect):
            self.add(elem)
        return self

    def __ixor__(self, s):
        # type: (Iterable[Any]) -> _BaseSet
        """
        Update the TerminalSet.
        Keep elements from self and other, but discard elements that are in both.
        :param other: Iterable object with elements to compare with.
        :return: Current instance with updated state.
        """
        return self.symmetric_difference_update(s)

    def discard(self, *elements):
        # type: (Any) -> None
        """
        Remove elements from the set.
        In case of terminals and nonterminals, remove rules using them as well.
        Unlike remove doesn't throw exception if the element is not inside the container.
        :param elements: Elements to remove
        """
        for el in elements:
            try:
                self.remove(el)
            except KeyError:
                continue

    def size(self):
        # type: () -> int
        """
        Get number of elements in the container.
        :return: Number of elements in the container.
        """
        return len(self)



#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 18.03.2019 18:17
:Licence MIT
Part of grammpy

"""
from typing import Callable, Any, Union, Generator
from grammpy import Rule, Nonterminal, Terminal


class Traversing:
    """
    Class that aggregates functions to traverse the parsed tree.
    """

    @staticmethod
    def traverse(root, callback, *args, **kwargs):
        # type: (Nonterminal, Callable[[Any, Callable, Any, Any], Generator], Any, Any) -> Generator
        """
        Traverse AST based on callback.
        :param root: Root element of the parsed tree.
        :param callback: Function that accepts current node, callback `c_2` and parameters from the parent.
        Function must yield individual values.
        Its possible to yield callback c_2 **call** on any node to call the recursion.
        The callback can accept parameters from the parent call.
        The root will receive parameters from the `traverse` call.

        Example of pre-order traversing:
        def traverse_func(item, callback):
            if isinstance(item, Rule):
                yield item
                for el in item.to_symbols:
                    yield callback(el)
            elif isinstance(item, Nonterminal):
                yield item
                yield callback(item.to_rule)
            else:
                yield item

        :return: Sequence of nodes to traverse.
        """
        class MyGenerator:
            def __init__(self, gen):
                self._gen = gen
            def __next__(self):
                return next(self._gen)

        def inner_callback(item, *args, **kwargs):
            return MyGenerator(callback(item, inner_callback, *args, **kwargs))

        to_call = list()
        to_call.append(inner_callback(root, *args, **kwargs))
        while len(to_call) > 0:
            current = to_call.pop()
            try:
                el = next(current)
                to_call.append(current)
                if isinstance(el, MyGenerator):
                    to_call.append(el)
                else:
                    yield el
            except StopIteration:
                continue

    @staticmethod
    def traverse_separated(root, callbackRules, callbackNonterminals, callbackTerminals, *args, **kwargs):
        # type: (Nonterminal, Callable[[Rule, Callable, Any, Any], Generator], Callable[[Nonterminal, Callable, Any, Any], Generator], Callable[[Terminal, Callable, Any, Any], Generator], Any, Any) -> Generator
        """
        Same as traverse method, but have different callbacks for rules, nonterminals and terminals.
        Functions accepts current node, callback `c_2` and parameters from the parent.
        Functions must yield individual values.
        Its possible to yield callback c_2 **call** on any node to call the recursion.
        The callback can accept parameters from the parent call.
        The root will receive parameters from the `traverse_separated` call.
        :param root: Root node of the parsed tree.
        :param callbackRules: Function to call for every rule.
        :param callbackNonterminals: Function to call for every nonterminal.
        :param callbackTerminals: Function to call for every terminal.
        :return: Sequence of nodes to traverse.
        """

        def separate_traverse_callback(item, callback, *args, **kwargs):
            if isinstance(item, Rule):
                return callbackRules(item, callback, *args, **kwargs)
            if isinstance(item, Nonterminal):
                return callbackNonterminals(item, callback, *args, **kwargs)
            if isinstance(item, Terminal):
                return callbackTerminals(item, callback, *args, **kwargs)

        return Traversing.traverse(root, separate_traverse_callback, *args, **kwargs)

    @staticmethod
    def pre_order(root):
        # type: (Nonterminal) -> Generator
        """
        Perform pre-order traversing. Expects tree like structure.
        Traverse in DFS fashion.
        :param root: Root tree of the parsed tree.
        :return: Sequence of nodes to traverse.
        """

        def traverse_rule(item, callback):
            yield item
            for el in item.to_symbols:
                yield callback(el)

        def traverse_nonterminal(item, callback):
            yield item
            yield callback(item.to_rule)

        def traverse_terminal(item, callback):
            yield item

        return Traversing.traverse_separated(root, traverse_rule, traverse_nonterminal, traverse_terminal)

    @staticmethod
    def post_order(root):
        # type: (Nonterminal) -> Generator
        """
        Perform post-order traversing. Expects tree like structure.
        Traverse in DFS fashion.
        :param root: Root node of the parsed tree.
        :return: Sequence of nodes to traverse.
        """

        def traverse_rule(item, callback):
            for el in item.to_symbols:
                yield callback(el)
            yield item

        def traverse_nonterminal(item, callback):
            yield callback(item.to_rule)
            yield item

        def traverse_terminal(item, callback):
            yield item

        return Traversing.traverse_separated(root, traverse_rule, traverse_nonterminal, traverse_terminal)

    @staticmethod
    def print(root):
        # type: (Union[Nonterminal,Terminal,Rule])-> str
        """
        Transform the parsed tree to the string. Expects tree like structure.
        You can see example output below.

        (R)SplitRules26
        |--(N)Iterate
        |  `--(R)SplitRules30
        |     `--(N)Symb
        |        `--(R)SplitRules4
        |           `--(T)e
        `--(N)Concat
           `--(R)SplitRules27
              `--(N)Iterate
                 `--(R)SplitRules30
                    `--(N)Symb
                       `--(R)SplitRules5
                          `--(T)f

        :param root: Root node of the parsed tree.
        :return: String representing the parsed tree (ends with newline).
        """
        # print the part before the element
        def print_before(previous=0, defined=None, is_last=False):
            defined = defined or {}
            ret = ''
            if previous != 0:
                for i in range(previous - 1):
                    # if the column is still active write |
                    if i in defined:
                        ret += '|  '
                    # otherwise just print space
                    else:
                        ret += '   '
                # if is current element last child, don't print |-- but `-- instead
                ret += '`--' if is_last else '|--'
            return ret

        # print the terminal
        def terminal_traverse(term, callback, previous=0, defined=None, is_last=False):
            before = print_before(previous, defined, is_last)
            yield before + '(T)' + str(term.s) + '\n'

        # print the nonterminal
        def nonterminal_traverse(nonterm, callback, previous=0, defined=None, is_last=False):
            before = print_before(previous, defined, is_last)
            yield before + '(N)' + nonterm.__class__.__name__ + '\n'
            yield callback(nonterm.to_rule, previous + 1, defined, True)

        # print the rule
        def rule_traverse(rule, callback, previous=0, defined=None, is_last=False):
            # print the rule name
            before = print_before(previous, defined, is_last)
            yield before + '(R)' + rule.__class__.__name__ + '\n'
            # register new column
            defined = defined or set()
            defined.add(previous)
            # print all childs except the last one
            for i in range(len(rule.to_symbols) - 1):
                yield callback(rule.to_symbols[i], previous + 1, defined, False)
            # unregister the column as last child print it automatically
            defined.remove(previous)
            yield callback(rule.to_symbols[-1], previous + 1, defined, True)

        res = Traversing.traverse_separated(root, rule_traverse, nonterminal_traverse, terminal_traverse)
        return str.join("", res)



#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 01.08.2017 08:46
:Licence MIT
Part of grammpy

"""

from .RuleException import RuleException


class RuleNotDefinedException(RuleException):
    """
    For Rule class the rule is not defined
    """
    def __init__(self, rule):
        self.rule = rule



#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 03.08.2017 10:04
:Licence MIT
Part of grammpy

"""

from .GrammpyException import GrammpyException


class NotRuleException(GrammpyException, TypeError):
    """
    Passed something else than Rule class
    """
    def __init__(self, rule):
        super().__init__()
        self.object = rule



#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 15.10.2018 15:10
:Licence MIT
Part of grammpy

"""
import inspect
from typing import Iterable, TYPE_CHECKING, Type

from ._BaseSet import _BaseSet
from ..Nonterminal import Nonterminal
from ...exceptions import NotNonterminalException

if TYPE_CHECKING:  # pragma: no cover
    from .. import Grammar


class _NonterminalSet(_BaseSet):
    """
    Set that store nonterminals inside the grammar.
    """

    def __init__(self, grammar, assign_map, iterable=None):
        # type: (Grammar, dict, Iterable[Type[Nonterminal]]) -> None
        """
        Create new instance of _NonterminalSet.
        :param grammar: Grammar for which create the set.
        :param assign_map: Map used for assignment rules to nonterminals.
        :param iterable: Nonterminals to insert.
        """
        self._grammar = grammar
        self._assign_map = assign_map
        super().__init__()
        self.add(*(iterable or []))

    @staticmethod
    def _control_nonterminal(nonterm):
        # type: (Type[Nonterminal]) -> None
        """
        Check if the nonterminal is valid.
        :param nonterm: Nonterminal to check.
        :raise NotNonterminalException: If the object doesn't inherit from Nonterminal class.
        """
        if not inspect.isclass(nonterm) or not issubclass(nonterm, Nonterminal):
            raise NotNonterminalException(nonterm)

    def add(self, *nonterminals):
        # type: (Iterable[Type[Nonterminal]]) -> None
        """
        Add nonterminals into the set.
        :param nonterminals: Nonterminals to insert.
        :raise NotNonterminalException: If the object doesn't inherit from Nonterminal class.
        """
        for nonterm in nonterminals:
            if nonterm in self:
                continue
            _NonterminalSet._control_nonterminal(nonterm)
            super().add(nonterm)
            self._assign_map[nonterm] = set()

    def remove(self, *nonterminals):
        # type: (Iterable[Type[Nonterminal]]) -> None
        """
        Remove nonterminals from the set.
        Removes also rules using this nonterminal.
        Set start symbol to None if deleting nonterminal is start symbol at the same time.
        :param nonterminals: Nonterminals to remove.
        """
        for nonterm in set(nonterminals):
            _NonterminalSet._control_nonterminal(nonterm)
            if nonterm not in self:
                raise KeyError('Nonterminal ' + nonterm.__name__ + ' is not inside')
            self._grammar.rules.remove(*self._assign_map[nonterm], _validate=False)
            del self._assign_map[nonterm]
            if self._grammar.start is nonterm:
                del self._grammar.start
            super().remove(nonterm)

    def __contains__(self, o):
        # type: (Type[Nonterminal]) -> bool
        """
        Check, if is nonterminal in the set.
        :param o: Nonterminal to check.
        :return: True if the nonterminal is in the set, false otherwise.
        :raise NotNonterminalException: If the object doesn't inherit from Nonterminal class.
        """
        _NonterminalSet._control_nonterminal(o)
        return super().__contains__(o)



#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 24.01.2019 23:30
:Licence MIT
Part of grammpy

"""
from typing import Type, TYPE_CHECKING, List, Tuple

from ..Rule import Rule as _Rule

if TYPE_CHECKING:  # pragma: no cover
    from .. import Rule


class SplitRule(_Rule):
    """
    Represent rule, that originate from splitting other rule.
    This mean that the second Rule class consists of multiple rules.
    """

    @staticmethod
    def _create_class(rule, index):
        # type: (Type[Rule], int) -> Type[SplitRule]
        """
        Create subtype of SplitRule based on rule.
        :param rule: Rule from which the SplitRule derive.
        :param index: Index of the rule (in original Rule class) to use for SplitRule.
        :return: Class inherited from SplitRule representing rule at index.
        """
        name = 'SplitRule[' + rule.__name__ + ';' + str(index) + ']'
        created = type(name, (SplitRule,), SplitRule.__dict__.copy())  # type: Type[SplitRule]
        created.rule = rule.rules[index]
        created.rule_index = index
        created.from_rule = rule
        return created

    from_rule = None  # type: Type[Rule]
    rule_index = None  # type: int
    rule = None  # type: Tuple[List[object], List[object]]



#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 24.01.2019 13:04
:Licence MIT
Part of grammpy

"""

from .Grammar import Grammar
# noinspection PyUnresolvedReferences
from .. import Nonterminal, Terminal, Rule, EPS, EPSILON



#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 23.12.2017 16:03
:Licence MIT
Part of grammpy

"""
from typing import TYPE_CHECKING

from .SplittedRules import splitted_rules

if TYPE_CHECKING:  # pragma no cover
    from .. import Nonterminal


class InverseCommon:
    """
    Class that aggregates functions transforming common parsed tree.
    """

    @staticmethod
    def splitted_rules(root):
        # type: (Nonterminal) -> Nonterminal
        """
        Replace SplittedRules in the parsed tree with the original one.
        This method is mandatory if you insert Rule class with multiple rules into the grammar.
        :param root: Root of the parsed tree.
        :return: Modified tree.
        """
        # TODO should by done automatically
        return splitted_rules(root)



#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 03.09.2017 10:49
:Licence MIT
Part of grammpy

"""
from typing import TYPE_CHECKING

from .ChomskyForm import transform_from_chomsky_normal_form
from .EpsilonRulesRemove import epsilon_rules_restore
from .UnitRulesRemove import unit_rules_restore

if TYPE_CHECKING:  # pragma: no cover
    from .. import Nonterminal

__all__ = ['InverseContextFree']


class InverseContextFree:
    """
    Class that aggregates functions transforming Context-Free parsed tree.
    """

    @staticmethod
    def unit_rules_restore(root):
        # type: (Nonterminal) -> Nonterminal
        """
        Transform parsed tree for grammar with removed unit rules.
        The unit rules will be returned back to the tree.
        :param root: Root of the parsed tree.
        :return: Modified tree.
        """
        return unit_rules_restore(root)

    @staticmethod
    def epsilon_rules_restore(root):
        # type: (Nonterminal) -> Nonterminal
        """
        Transform parsed tree to contain epsilon rules originally removed from the grammar.
        :param root: Root of the parsed tree.
        :return: Modified tree including epsilon rules.
        """
        return epsilon_rules_restore(root)

    @staticmethod
    def transform_from_chomsky_normal_form(root):
        # type: (Nonterminal) -> Nonterminal
        """
        Transform the tree created by grammar in the Chomsky Normal Form to original rules.
        :param root: Root of parsed tree.
        :return: Modified tree.
        """
        return transform_from_chomsky_normal_form(root)

    @staticmethod
    def reverse_cyk_transforms(root):
        # type: (Nonterminal) -> Nonterminal
        """
        Reverse transformation made to grammar before CYK.
        Performs following steps:
        - transform from chomsky normal form
        - restore unit rules
        - restore epsilon rules
        :param root: Root node of the parsed tree.
        :return: Restored parsed tree.
        """
        root = InverseContextFree.transform_from_chomsky_normal_form(root)
        root = InverseContextFree.unit_rules_restore(root)
        root = InverseContextFree.epsilon_rules_restore(root)
        return root



#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 10.08.2017 20:33
:Licence MIT
Part of grammpy

"""
from copy import copy
from typing import TYPE_CHECKING

from ... import EPSILON

if TYPE_CHECKING:  # pragma: no cover
    from ... import Grammar


def remove_nongenerating_nonterminals(grammar, inplace=False):
    # type: (Grammar, bool) -> Grammar
    """
    Remove nongenerating symbols from the grammar.
    Nongenerating symbols are symbols, that don't generate sequence of terminals.
    For example never ending recursion.
    :param grammar: Grammar where to remove nongenerating symbols.
    :param inplace: True if transformation should be performed in place. False by default.
    :return: Grammar without nongenerating symbols.
    """
    # copy if required
    if inplace is False:
        grammar = copy(grammar)
    # create working sets
    generates = grammar.terminals.copy()
    generates.add(EPSILON)
    rules = grammar.rules.copy()
    # iterate until the set doesn't change
    while True:
        # create set for the next iteration
        additional = generates.copy()
        # iterate over unprocessed rules
        for rule in rules.copy():
            rightPart = rule.right
            allIn = True
            # check if all symbols on the right part of rule are in generates set
            for symbol in rightPart:
                if symbol not in generates:
                    allIn = False
                    break
            # Symbol is missing so rule is not process
            if not allIn:
                continue
            # Rule is process - remove it from processing rules and make symbol as generating
            additional.add(rule.fromSymbol)
            rules.remove(rule)
            # end of rules iterations
        # ff current and previous iterations are same, than end iterations
        if additional == generates:
            break
        # swap sets from previous and current iterations
        generates = additional
    # remove nonterms that are not generating
    nongenerating = grammar.nonterminals.difference(generates)
    grammar.nonterminals.remove(*nongenerating)
    # return the grammar
    return grammar



#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 02.02.2019 20:17
:Licence MIT
Part of grammpy

"""

from typing import TYPE_CHECKING, List, Type

if TYPE_CHECKING:  # pragma: no cover
    from .Grammar import Grammar
    from .. import Rule


class RulesClass:
    """
    Class that help old API Grammar to distinguish between call of rules method and access to rule property.
    """

    def __init__(self, gr):
        # type: (Grammar) -> None
        """
        Return new instance of RulesClass.
        This class is only for internal use.
        :param gr: Grammar to reference.
        """
        self._gr = gr
        super().__init__()

    def __call__(self):
        # type: () -> List[Type[Rule]]
        """
        Get all rules within the grammar.
        :return: List of rules in the grammar.
        """
        return self._gr._rules()

    def __iter__(self):
        return self._gr._gr.rules.__iter__()

    def __getattribute__(self, item):
        if item in {'__call__', '_gr'}:
            return object.__getattribute__(self, item)
        return getattr(self._gr._gr.rules, item)



#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 10.08.2017 20:32
:Licence MIT
Part of grammpy

"""

from .find_symbols_reachable_by_unit_rules import find_nonterminals_reachable_by_unit_rules, UnitSymbolReachability
from .remove_unit_rules import remove_unit_rules, ReducedUnitRule
from .unit_rules_restore import unit_rules_restore



#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 23.06.2017 16:45
:Licence MIT
Part of grammpy

"""

from typing import Any, List, TYPE_CHECKING

from .support._MetaRule import _MetaRule
from .support._WeakList import _WeakList

if TYPE_CHECKING:  # pragma: no cover
    from .support._RuleConnectable import _RuleConnectable


class Rule(metaclass=_MetaRule):
    """
    Basic implementation of rules.
    For definition of rule, you can use following attributes:
    fromSymbol = EPSILON
    toSymbol = EPSILON
    left = [EPSILON]
    right = [EPSILON]
    rule = ([EPSILON], [EPSILON])
    rules = [([EPSILON], [EPSILON])]
    """

    def __init__(self):
        # type: () -> Rule
        """
        Create new instance of rule.
        """
        self._from_symbols = _WeakList()
        self._to_symbols = list()

    def __getattr__(self, name):
        # type: (str) -> Any
        """
        Handle special attributes: toSymbol fromSymbol left right rule rules.
        For the rest throws AttributeError.
        :param name: Name of the attribute to find.
        :throw AttributeError:
        """
        if name in {'toSymbol',
                    'fromSymbol',
                    'left',
                    'right',
                    'rule',
                    'rules'}:
            return getattr(self.__class__, name)
        raise AttributeError(self.__class__.__name__ + ' doesn\'t have attribute ' + name)

    @property
    def from_symbols(self):
        # type: () -> List[_RuleConnectable]
        """
        Instances that the rule derived from.
        :return: List of symbols.
        """
        return list(self._from_symbols)

    @property
    def to_symbols(self):
        # type: () -> List[_RuleConnectable]
        """
        Instances that the rule derived to.
        :return: List of symbols.
        """
        return self._to_symbols



#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 03.09.2017 10:48
:Licence MIT
Part of grammpy

"""
from typing import TYPE_CHECKING

from .remove_unit_rules import ReducedUnitRule
from ..Traversing import Traversing

if TYPE_CHECKING:  # pragma: no cover
    from ... import Nonterminal, Rule


def unit_rules_restore(root):
    # type: (Nonterminal) -> Nonterminal
    """
    Transform parsed tree for grammar with removed unit rules.
    The unit rules will be returned back to the tree.
    :param root: Root of the parsed tree.
    :return: Modified tree.
    """
    items = Traversing.post_order(root)
    items = filter(lambda x: isinstance(x, ReducedUnitRule), items)
    for rule in items:
        parent_nonterm = rule.from_symbols[0]  # type: Nonterminal
        # restore chain of unit rules
        for r in rule.by_rules:
            created_rule = r()  # type: Rule
            parent_nonterm._set_to_rule(created_rule)
            created_rule._from_symbols.append(parent_nonterm)
            created_nonterm = r.toSymbol()  # type: Nonterminal
            created_rule._to_symbols.append(created_nonterm)
            created_nonterm._set_from_rule(created_rule)
            parent_nonterm = created_nonterm
        # restore last rule
        last_rule = rule.end_rule()  # type: Rule
        last_rule._from_symbols.append(parent_nonterm)
        parent_nonterm._set_to_rule(last_rule)
        for ch in rule.to_symbols:  # type: Nonterminal
            ch._set_from_rule(last_rule)
            last_rule._to_symbols.append(ch)
    return root



#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 31.08.2017 14:47
:Licence MIT
Part of grammpy

"""

from typing import Iterable, TYPE_CHECKING, Generator, Any, Dict, Set, Type

from .Field import Field
from .PlaceItem import PlaceItem
from ... import *
from ...exceptions import NotParsedException, StartSymbolNotSetException

if TYPE_CHECKING:  # pragma: no cover
    from ... import Grammar


def _create_mapping(grammar):
    # type: (Grammar) -> (Dict[int, Set[Type[Rule]]], Dict[int, Set[Type[Rule]]])
    """
    Create mapping between symbols and rules rewritable to these symbols.
    :param grammar: Grammar to use.
    :return: Tuple of two dictionaries.
    First dictionary maps rule to terminal hash.
    Second dictionary maps rule to nonterminals hash.
    """
    termmap = dict()
    rulemap = dict()
    for r in grammar.rules:
        if len(r.right) == 1:
            # rule to terminal
            h = hash(r.toSymbol)
            if h not in termmap:
                termmap[h] = set()
            termmap[h].add(r)
        else:
            # rule to two nonterms
            key = hash(tuple(r.right))
            if key not in rulemap:
                rulemap[key] = set()
            rulemap[key].add(r)
    return (termmap, rulemap)

def _all_combinations(tpl):
    # type: ((Iterable, Iterable)) -> Generator[tuple]
    """
    Return all combinations of iterables passed as a parameter.
    :param tpl: Tuple of two iterable.
    :return: Generator of tuples.
    """
    for f in tpl[0]:
        for s in tpl[1]:
            yield (f, s)


def cyk(grammar, parse_sequence):
    # type: (Grammar, Iterable[Any]) -> Nonterminal
    """
    Perform CYK algorithm.
    :param grammar: Grammar to use in Chomsky Normal Form.
    :param parse_sequence: Input sequence to parse.
    :return: Instance of root Nonterminal in parsed tree.
    """
    # check start symbol
    if grammar.start is None:
        raise StartSymbolNotSetException()
    # create variables
    parse_sequence = list(parse_sequence)
    input_length = len(parse_sequence)
    index = input_length - 1
    f = Field(input_length)
    # creating mapping for speedup rules searching
    (termmap, rulemap) = _create_mapping(grammar)
    # fill first line with rules directly rewritable to terminal
    f.fill(termmap, parse_sequence)
    # fill rest of fields
    for y in range(1, input_length):
        for x in range(input_length - y):
            positions = f.positions(x, y)
            pairs_of_rules = [(f.rules(pos[0].x, pos[0].y),
                      f.rules(pos[1].x, pos[1].y))
                     for pos in positions]
            rules = set()
            for pair_of_rule in pairs_of_rules:
                for (first_rule, second_rule) in _all_combinations(pair_of_rule):
                    h = hash((first_rule.fromSymbol, second_rule.fromSymbol))
                    if h in rulemap:
                        for r in rulemap[h]: # list of rules
                            rules.add(PlaceItem(r, first_rule, second_rule))
            f.put(x, y, list(rules))
    # Check if is start symol on the bottom of field
    if grammar.start not in [r.fromSymbol for r in f.rules(0, input_length - 1)]:
        raise NotParsedException()
    # Find init symbol and rule
    start = grammar.start()  # type: Nonterminal
    start_rule = [r for r in f.rules(0, input_length - 1) if grammar.start == r.fromSymbol][0]
    # Prepare buffer for proccess
    to_process = list()
    to_process.append({'n': start, 'r': start_rule})
    # Prepare tree
    while len(to_process) > 0:
        working = to_process.pop()
        rule_class = working['r']
        working_nonterm = working['n']  # type: Nonterminal
        # its middle rule - not rewritable to nonterminal
        if isinstance(rule_class, PlaceItem):
            created_rule = rule_class.rule()  # type: Rule
            working_nonterm._set_to_rule(created_rule)
            created_rule._from_symbols.append(working_nonterm)
            for side in rule_class.to_rule:
                symbol = side.fromSymbol()  # type: Nonterminal
                symbol._set_from_rule(created_rule)
                created_rule._to_symbols.append(symbol)
                to_process.append({'n': symbol, 'r': side})
        # it is rule rewritable to nonterminal
        else:
            created_rule = rule_class()  # type: Rule
            working_nonterm._set_to_rule(created_rule)
            created_rule._from_symbols.append(working_nonterm)
            t = Terminal(parse_sequence[index])
            index -= 1
            created_rule._to_symbols.append(t)
            t._set_from_rule(created_rule)
    return start



#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 31.08.2017 12:11
:Licence MIT
Part of grammpy


Original implementation: https://github.com/apieum/weakreflist
Changes: List raise TreeDeletedException instead of reducing its size
"""

import weakref
from typing import List, Optional, Any, Iterator, Iterable, Callable

from ...exceptions import TreeDeletedException


class _WeakList(list):
    """
    Class that behave like list, but store just weakref of items.
    """

    def __init__(self, items=None):
        # type: (Optional[List[object]]) -> _WeakList
        """
        Create instance of _Weaklist.
        :param items: Optional iterable object with values to add.
        """
        list.__init__(self, self._refs(items or []))

    def value(self, item):
        # type: (Any) -> Any
        """
        Return value stored in weakref.
        :param item: Object from which get the value.
        :return: Value stored in the weakref, otherwise original value.
        :raise TreeDeletedException: when weakref is already deleted.
        """
        if isinstance(item, weakref.ReferenceType):
            if item() is None:
                raise TreeDeletedException()
            return item()
        return item

    def ref(self, item):
        # type: (Any) -> weakref.ReferenceType
        """
        Create weakref for the parameter.
        :param item: Object from which make the weak ref.
        :return: Weakref of parameter, parameter if exception occurs.
        """
        try:
            item = weakref.ref(item)
        finally:
            return item

    def __contains__(self, item):
        # type(Any) -> bool
        """
        Check if is parameter in WeakList.
        :param item: Object to check the existence.
        :return: True if is parameter in WeakList, false otherwise.
        """
        return list.__contains__(self, self.ref(item))

    def __getitem__(self, index):
        # type: (int) -> Any
        """
        Return item on index.
        :param index: Index of item.
        :return: Element at specific index.
        """
        items = list.__getitem__(self, index)
        return type(self)(self._values(items)) if isinstance(index, slice) else self.value(items)

    def __setitem__(self, index, item):
        # type: (int, Any) -> None
        """
        Set value at specific index.
        :param index: Index where to store item.
        :param item: Item to store.
        """
        items = self._refs(item) if isinstance(index, slice) else self.ref(item)
        return list.__setitem__(self, index, items)

    def __iter__(self):
        # type: () -> Iterator[Any]
        """
        Get iterator fot WeakList.
        :return: Iterator of the list.
        """
        return iter(self[index] for index in range(len(self)))

    def __reversed__(self):
        # type: () -> _WeakList
        """
        Reverse items in WeakList.
        The original list is unmodified.
        :return: Reversed list.
        """
        reversed_self = type(self)(self)
        reversed_self.reverse()
        return reversed_self

    def append(self, item):
        # type: (Any) -> None
        """
        Add parameter to the end of the WeakList.
        :param item: Item to add.
        """
        list.append(self, self.ref(item))

    def remove(self, item):
        # type: (Any) -> None
        """
        Remove first occurrence of the parameter.
        :param item: Value to delete from the WeakList.
        """
        return list.remove(self, self.ref(item))

    def remove_all(self, item):
        # type: (Any) -> None
        """
        Remove all occurrence of the parameter.
        :param item: Value to delete from the WeakList.
        """
        item = self.ref(item)
        while list.__contains__(self, item):
            list.remove(self, item)

    def index(self, item, **kwargs):
        # type: (Any, dict) -> int
        """
        Get index of the parameter.
        :param item: Item for which get the index.
        :return: Index of the parameter in the WeakList.
        """
        return list.index(self, self.ref(item), **kwargs)

    def count(self, item):
        # type: (Any) -> int
        """
        Get count of items that match the parameter.
        :param item: Item for which search occurrence.
        :return: Count of items in the WeakList.
        """
        return list.count(self, self.ref(item))

    def pop(self, index=-1):
        # type: (int) -> Any
        """
        Delete item at specific position (last by default) and return it.
        :param index: Index where to delete the item, last item by default.
        :return: Deleted item.
        """
        return self.value(list.pop(self, index))

    def insert(self, index, item):
        # type: (int, Any) -> None
        """
        Insert item at the specific index.
        :param index: Index where to insert the item.
        :param item: Item to insert.
        """
        return list.insert(self, index, self.ref(item))

    def extend(self, items):
        # type: (Iterable[Any]) -> None
        """
        Add items to the end of the WeakList.
        :param items: Iterable object of items to insert.
        """
        return list.extend(self, self._refs(items))

    def __iadd__(self, other):
        # type: (Iterable[Any]) -> _WeakList
        """
        Join WeakList with another iterable.
        :param other: Iterable to join with.
        """
        return list.__iadd__(self, self._refs(other))

    def _refs(self, items):
        # type: (Iterable[Any]) -> Iterable[weakref.ReferenceType]
        """
        Call ref on every item in the parameter.
        :param items: Iterable of items.
        :return: Iterable of weakrefs.
        """
        return map(self.ref, items)

    def _values(self, items):
        # type: (Iterable[weakref.ReferenceType]) -> Iterable[Any]
        """
        Call value on every item.
        :param items: Iterable of items.
        :return: Iterable of values.
        """
        return map(self.value, items)

    def _sort_key(self, key=None):
        return self.value if key is None else lambda item: key(self.value(item))

    def sort(self, *, key: Optional[Callable[[Any], Any]] = None, reverse: bool = False) -> None:
        """
        Sort _WeakList.
        :param key: Key by which to sort, default None.
        :param reverse: True if return reversed WeakList, false by default.
        """
        return list.sort(self, key=self._sort_key(key), reverse=reverse)



#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 28.04.2018 19:06
:Licence MIT
Part of grammpy

"""

from .GrammpyException import GrammpyException

class ParsingException(GrammpyException):
    """
    Represent error with parsing
    """
    pass


#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 24.12.2017 11:31
:Licence MIT
Part of grammpy

"""
from typing import TYPE_CHECKING, Union
from .. import Nonterminal, Terminal, Rule

if TYPE_CHECKING:  # pragma: no cover
    from ..representation.support._RuleConnectable import _RuleConnectable


class Manipulations:
    """
    Class that aggregate functions modifying the parsed tree.
    """

    @staticmethod
    def replaceRule(oldRule, newRule):
        # type: (Rule, Rule) -> Rule
        """
        Replace instance of Rule with another one.
        :param oldRule: Instance in the tree.
        :param newRule: Instance to replace with.
        :return: New instance attached to the tree.
        """
        for par in oldRule.from_symbols:
            par._set_to_rule(newRule)
            newRule._from_symbols.append(par)
        for ch in oldRule.to_symbols:
            ch._set_from_rule(newRule)
            newRule._to_symbols.append(ch)
        return newRule

    @staticmethod
    def replaceNode(oldNode, newNode):
        # type: (_RuleConnectable, _RuleConnectable) -> _RuleConnectable
        """
        Replace instance of Nonterminal or Terminal in the tree with another one.
        :param oldNode: Old nonterminal or terminal already in the tree.
        :param newNode: Instance of nonterminal or terminal to replace with.
        :return: Instance `newNode` in the tree.
        """
        if oldNode.from_rule is not None and len(oldNode.from_rule.to_symbols) > 0:
            indexParent = oldNode.from_rule.to_symbols.index(oldNode)
            oldNode.from_rule.to_symbols[indexParent] = newNode
            newNode._set_from_rule(oldNode.from_rule)
        if oldNode.to_rule is not None and len(oldNode.to_rule.from_symbols) > 0:
            indexChild = oldNode.to_rule.from_symbols.index(oldNode)
            oldNode.to_rule._from_symbols[indexChild] = newNode
            newNode._set_to_rule(oldNode.to_rule)
        return newNode

    @staticmethod
    def replace(oldEl, newEl):
        # type: (Union[Rule, _RuleConnectable], Union[Rule, _RuleConnectable]) -> Union[Rule, _RuleConnectable]
        """
        Replace element in the parsed tree. Can be nonterminal, terminal or rule.
        :param oldEl: Element already in the tree.
        :param newEl: Element to replace with.
        :return: New element attached to the tree.
        """
        if isinstance(oldEl, Rule):
            return Manipulations.replaceRule(oldEl, newEl)
        if isinstance(oldEl, (Nonterminal, Terminal)):
            return Manipulations.replaceNode(oldEl, newEl)



#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 10.08.2017 20:32
:Licence MIT
Part of grammpy

"""

from .ContextFree import ContextFree
from .InverseCommon import InverseCommon
from .InverseContextFree import InverseContextFree
from .Manipulations import Manipulations
from .Traversing import Traversing



#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 15.10.2018 15:10
:Licence MIT
Part of grammpy

"""
import inspect
from typing import Iterable, List, TYPE_CHECKING, Type, Generator, Optional

from deprecated import deprecated

from .SplitRule import SplitRule
from ._BaseSet import _BaseSet
from .. import Rule
from ...exceptions import NotRuleException, NonterminalDoesNotExistsException, TerminalDoesNotExistsException

if TYPE_CHECKING:  # pragma: no cover
    from .. import Grammar


class _RulesSet(_BaseSet):
    """
    Set that store terminals inside the grammar.
    """

    def __init__(self, grammar, assign_map, iterable=None):
        # type: (Grammar, dict, Iterable[Type[Rule]]) -> None
        """
        Create new instance of _RulesSet.
        :param grammar: Grammar for which create the set.
        :param assign_map: Map used for assignment rules to terminals or nonterminals.
        This map will be used for deleting rules with terminals or nonterminals within.
        :param iterable: Terminals to insert.
        """
        self._grammar = grammar
        self._assign_map = assign_map
        super().__init__()
        self.add(*(iterable or []))

    def _validate_rule(self, rule):
        # type: (Type[Rule]) -> None
        """
        Validate rule. Valid rule must inherit from Rule and have valid syntax.
        :param rule: Rule to validate.
        :raise NotRuleException: If the parameter doesn't inherit from Rule.
        """
        if not inspect.isclass(rule) or not issubclass(rule, Rule):
            raise NotRuleException(rule)
        rule.validate(self._grammar)

    def _split_rules(self, original_rule):
        # type: (Type[Rule]) -> Iterable[Type[Rule]]
        """
        Splits Rule class with multiple rules into separate classes, each with only one rule defined.
        The created rules inherits from SplitRule.
        If parameter define only one rule, its not split.
        :param original_rule: Rule to split.
        :return: Iterable of rules derived from the parameter.
        """
        if original_rule.count == 1:
            return [original_rule]

        def yielding(original_rule):
            for rule_index in range(original_rule.count):
                yield SplitRule._create_class(original_rule, rule_index)

        return yielding(original_rule)

    def add(self, *rules):
        # type: (Iterable[Type[Rule]]) -> List[Type[Rule]]
        """
        Add rules into the set. Each rule is validated and split if needed.
        :param rules: Rules to insert.
        :return: Inserted rules.
        :raise NotRuleException: If the parameter doesn't inherit from Rule.
        :raise RuleException: If the syntax of the rule is invalid.
        """
        return list(self._add(*rules))

    def _add(self, *rules):
        # type: (Iterable[Type[Rule]]) -> Generator[Type[Rule]]
        """
        Add rules into the set. Each rule is validated and split if needed.
        The method add the rules into dictionary, so the rule can be deleted with terminals or nonterminals.
        :param rules: Rules to insert.
        :return: Inserted rules.
        :raise NotRuleException: If the parameter doesn't inherit from Rule.
        :raise RuleException: If the syntax of the rule is invalid.
        """
        for rule in rules:
            if rule in self:
                continue
            self._validate_rule(rule)
        for rule in rules:
            for r in self._split_rules(rule):
                for side in r.rule:
                    for s in side:
                        self._assign_map[s].add(r)
                super().add(r)
                yield r

    def remove(self, *rules, _validate=True):
        # type: (Iterable[Type[Rule]], bool) -> None
        """
        Remove rules from the set.
        :param rules: Rules to remove.
        :param _validate: True if the rule should be validated before deleting.
        This parameter is only for internal use.
        :raise NotRuleException: If the parameter doesn't inherit from Rule.
        :raise RuleException: If the syntax of the rule is invalid.
        :raise KeyError: If the rule is not in the grammar.
        """
        all_rules = set()
        for rule in rules:
            if _validate:
                self._validate_rule(rule)
            for r in self._split_rules(rule):
                if not self.__contains__(rule, _validate=False):
                    raise KeyError('Rule ' + rule.__name__ + ' is not inside')
                all_rules.add(r)
        for rule in all_rules:
            for side in rule.rule:
                for s in side:
                    self._assign_map[s].discard(rule)
            super().remove(rule)

    def __contains__(self, o, *, _validate=True):
        # type: (Type[Rule], _, bool) -> bool
        """
        Check, if is rule in the set.
        :param o: Rule to check.
        :return: True if the rule is in the set, false otherwise.
        :raise NotRuleException: If the parameter doesn't inherit from Rule.
        :raise RuleException: If the syntax of the rule is invalid.
        """
        try:
            if _validate:
                self._validate_rule(o)
            for r in self._split_rules(o):
                if not super().__contains__(r):
                    return False
            return True
        except (TerminalDoesNotExistsException, NonterminalDoesNotExistsException):
            return False

    @deprecated(version='2.0.0', reason='Only for purpose of old API')
    def get(self, *rules):
        # type: (Iterable[Type[Rule]]) -> List[Type[Rule]]
        """
        Get rules representing parameters.
        The return rules can be different from parameters, in case parameter define multiple rules in one class.
        :param rules: For which rules get the representation.
        :return: List of rules representing parameters.
        :raise NotRuleException: If the parameter doesn't inherit from Rule.
        :raise RuleException: If the syntax of the rule is invalid.
        """
        return list(self._get(*rules))

    def _get(self, *rules):
        # type: (Iterable[Type[Rule]]) -> Generator[Type[Rule]]
        """
        Get rules representing parameters.
        The return rules can be different from parameters, in case parameter define multiple rules in one class.
        :param rules: For which rules get the representation.
        :return: List of rules representing parameters.
        :raise NotRuleException: If the parameter doesn't inherit from Rule.
        :raise RuleException: If the syntax of the rule is invalid.
        """
        for rule in rules:
            if not inspect.isclass(rule) or not issubclass(rule, Rule):
                raise NotRuleException(rule)
            for r in self._split_rules(rule):
                yield self._find_rule(r)

    def _find_rule(self, rule):
        # type: (Type[Rule]) -> Optional[Type[Rule]]
        """
        Find rule representing parameter.
        :param rule: Rule to find.
        :return: Rule stored inside the set.
        """
        for r in self:
            if r == rule:
                return r
        return None



#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 28.07.2017 09:12
:Licence MIT
Part of grammpy

"""
from typing import Any

from .GrammpyException import GrammpyException


class NotNonterminalException(GrammpyException, TypeError):
    """
    Object is not Nonterminal class
    """
    def __init__(self, parameter, *args: Any) -> None:
        super().__init__(*args)
        self.object = parameter



#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 01.08.2017 08:54
:Licence MIT
Part of grammpy

"""

from .RuleException import RuleException


class CantCreateSingleRuleException(RuleException):
    """
    From Rule class with multiple rules defined library was unable to split separate rules
    """
    def __init__(self, rules):
        self.rules = rules



#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 02.08.2017 17:51
:Licence MIT
Part of grammpy

"""

from .RuleSyntaxException import RuleSyntaxException


class UselessEpsilonException(RuleSyntaxException):
    """
    Exception that exception used in grammar is useless
    """
    def __init__(self, rule):
        super().__init__(rule, 'Usage of epsilon in this context if useless')



#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 28.08.2017 10:19
:Licence MIT
Part of grammpy

"""
from copy import copy
from queue import Queue
from typing import Any, Type

from ... import *


class ChomskyNonterminal(Nonterminal):
    """
    Base class for all nonterminals used in Chomsky Normal Form.
    """
    pass


class ChomskyGroupNonterminal(ChomskyNonterminal):
    """
    Class representing nonterminal that represents nonterminal tail when rule is splitted.
    """
    group = []


class ChomskyTermNonterminal(ChomskyNonterminal):
    """
    Class representing nonterminal directly rewritable to terminal.
    """
    for_term = None


class ChomskyRule(Rule):
    """
    Base class for all rules used in Chomsky Normal Form.
    """
    pass


class ChomskySplitRule(ChomskyRule):
    """
    Class representing split rule with more than two nonterminals.
    First symbol of the right side is first symbol in the original rule.
    Second symbol of the right side is always ChomskyGroupNonterminal.
    Rule from the ChomskyGroupNonterminal is always ChomskyRestRule.
    """
    from_rule = None


class ChomskyRestRule(ChomskyRule):
    """
    Class representing rule for ChomskyGroupNonterminal class.
    Always rewrote from ChomskyGroupNonterminal.
    """
    from_rule = None


class ChomskyTerminalReplaceRule(ChomskyRule):
    """
    Replacement of rule, that have two symbol on the right side, but one of it is terminal.
    """
    from_rule = None
    replaced_index = None


class ChomskyTermRule(ChomskyRule):
    """
    Rule that directly rewrite nonterminal to terminal.
    """
    pass


class Container:
    """
    Represent container that hold terminal, appropriate ChomskyTermNonterminal for which exists rule
    to rewrite to that terminal and ChomskyTermRule, that rewrites the nonterminal to terminal.
    Purpose of this class is to keep them together.
    """

    def __init__(self, terminal, nonterminal, rule):
        # type: (Any, Type[ChomskyTermNonterminal], Type[ChomskyTermRule]) -> None
        """
        Create instance of the container.
        :param terminal: Original terminal.
        :param nonterminal: Nonterminal directly rewritable to terminal.
        :param rule: Rule that rewrite nonterminal to terminal.
        """
        self.used = False
        self.terminal = terminal
        self.nonterminal = nonterminal
        self.rule = rule


class TerminalsFilling:
    """
    Store all terminals and their appropriate ChomskyTermNonterminal and ChomskyTermRule.
    Automatically add rule and nonterminal into grammar if the rule is used.
    """

    def __init__(self, grammar):
        # type: (Grammar) -> None
        """
        :param grammar: Grammar to work with. Can be modified.
        """
        self._grammar = grammar
        self._items = dict()
        for term in grammar.terminals:
            created_nonterm = type("ChomskyTerm[" + str(term) + "]",
                                   (ChomskyTermNonterminal,),
                                   ChomskyTermNonterminal.__dict__.copy())  # type: Type[ChomskyTermNonterminal]
            created_nonterm.for_term = term
            created_rule = type("ChomskyTerm[" + str(term) + "]",
                                (ChomskyTermRule,),
                                ChomskyTermRule.__dict__.copy())  # type: Type[ChomskyTermRule]
            created_rule.rule = ([created_nonterm], [term])
            self._items[term] = Container(term, created_nonterm, created_rule)

    def get(self, term):
        # type: (Any) -> Type[ChomskyTermNonterminal]
        """
        Get nonterminal rewritable to term.
        If the rules is not in the grammar, nonterminal and rule rewritable to terminal are add into grammar.
        :param term: Term for which get the nonterminal.
        :return: ChomskyTermNonterminal class for terminal.
        """
        if self._items[term].used is False:
            cont = self._items[term]
            self._grammar.nonterminals.add(cont.nonterminal)
            self._grammar.rules.add(cont.rule)
            cont.used = True
        return self._items[term].nonterminal


def transform_to_chomsky_normal_form(grammar, inplace=False):
    # type: (Grammar, bool) -> Grammar
    """
    Transform grammar to Chomsky Normal Form.
    :param grammar: Grammar to transform.
    :param inplace: True if transformation should be performed in place. False by default.
    :return: Grammar in Chomsky Normal Form.
    """
    # Copy if required
    if inplace is False:
        grammar = copy(grammar)
    # Create nonterminals rewritable to the terminal.
    # They will be inserted into the grammar as needed.
    fill = TerminalsFilling(grammar)
    to_process = Queue()
    for r in grammar.rules:
        to_process.put(r)
    while not to_process.empty():
        rule = to_process.get()  # type: Type[Rule]
        # Check, if rule must be split
        if len(rule.right) > 2:
            grammar.rules.remove(rule)
            # create nonterm that represent group on the right
            created_nonterm = type("ChomskyGroup[" + rule.__name__ + "]",
                                   (ChomskyGroupNonterminal,),
                                   ChomskyGroupNonterminal.__dict__.copy())  # type: Type[ChomskyGroupNonterminal]
            created_nonterm.group = rule.right[1:]
            # create rule that replace current
            created_left_rule = type("ChomskySplit[" + rule.__name__ + "]",
                                     (ChomskySplitRule,),
                                     ChomskySplitRule.__dict__.copy())  # type: Type[ChomskySplitRule]
            created_left_rule.rule = ([rule.fromSymbol], [rule.right[0], created_nonterm])
            created_left_rule.from_rule = rule
            # create rule with symbols on the right
            created_right_rule = type("ChomskyRest[" + rule.__name__ + "]",
                                      (ChomskyRestRule,),
                                      ChomskyRestRule.__dict__.copy())  # type: Type[ChomskyRestRule]
            created_right_rule.rule = ([created_nonterm], rule.right[1:])
            created_right_rule.from_rule = rule
            # add it to the grammar
            grammar.nonterminals.add(created_nonterm)
            grammar.rules.add(created_left_rule, created_right_rule)
            to_process.put(created_left_rule)
            to_process.put(created_right_rule)
        # check, if must replace terminal
        elif len(rule.right) == 2:
            if rule.right[0] in grammar.terminals:
                # first symbol is terminal
                # remove rule from grammar
                grammar.rules.remove(rule)
                # get nonterminal rewritable to the terminal and add that rules into the grammar implicitly
                symb = fill.get(rule.right[0])
                # create rule replacing the original one
                created = type("ChomskyLeft[" + rule.__name__ + "]",
                               (ChomskyTerminalReplaceRule,),
                               ChomskyTerminalReplaceRule.__dict__.copy())  # type: Type[ChomskyTerminalReplaceRule]
                created.rule = ([rule.fromSymbol], [symb, rule.right[1]])
                created.from_rule = rule
                created.replaced_index = 0
                # add it into the grammar
                grammar.rules.add(created)
                to_process.put(created)
            elif rule.right[1] in grammar.terminals:
                # second symbol is terminal
                # remove rule from grammar
                grammar.rules.remove(rule)
                # get nonterminal rewritable to the terminal and add that rules into the grammar implicitly
                symb = fill.get(rule.right[1])
                # create rule replacing the original one
                created = type("ChomskyRight[" + rule.__name__ + "]",
                               (ChomskyTerminalReplaceRule,),
                               ChomskyTerminalReplaceRule.__dict__.copy())
                created.rule = ([rule.fromSymbol], [rule.right[0], symb])
                created.from_rule = rule
                created.replaced_index = 1
                # add it into the grammar
                grammar.rules.add(created)
                to_process.put(created)
    return grammar



#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 10.08.2017 20:33
:Licence MIT
Part of grammpy

"""

from typing import TYPE_CHECKING, Dict

from .ChomskyForm import *
from .EpsilonRulesRemove import *
from .NongeneratingSymbolsRemove import *
from .UnitRulesRemove import *
from .UnreachableSymbolsRemove import *

if TYPE_CHECKING:  # pragma: no cover
    from .. import Grammar

__all__ = ['ContextFree']


class ContextFree:
    """
    Class that aggregates functions transforming context-free grammars.
    """

    EpsilonRemovedRule = EpsilonRemovedRule
    ReducedUnitRule = ReducedUnitRule
    ChomskyRule = ChomskyRule
    ChomskySplitRule = ChomskySplitRule
    ChomskyTermRule = ChomskyTermRule
    ChomskyRestRule = ChomskyRestRule
    ChomskyTerminalReplaceRule = ChomskyTerminalReplaceRule

    ChomskyNonterminal = ChomskyNonterminal
    ChomskyGroupNonterminal = ChomskyGroupNonterminal
    ChomskyTermNonterminal = ChomskyTermNonterminal

    UnitSymbolRechablingResults = UnitSymbolReachability

    @staticmethod
    def remove_nongenerating_nonterminals(grammar, inplace=False):
        # type: (Grammar, bool) -> Grammar
        """
        Remove nongenerating symbols from the grammar.
        Nongenerating symbols are symbols, that don't generate sequence of terminals.
        For example never ending recursion.
        :param grammar: Grammar where to remove nongenerating symbols.
        :param inplace: True if transformation should be performed in place. False by default.
        :return: Grammar without nongenerating symbols.
        """
        return remove_nongenerating_nonterminals(grammar, inplace)

    @staticmethod
    def is_grammar_generating(grammar, remove=False):
        # type: (Grammar, bool) -> bool
        """
        Check if is grammar is generating.
        Generating grammar generates at least one sentence.
        :param grammar: Grammar to check.
        :param remove: True to remove nongenerating symbols from the grammar.
        :return: True if is grammar generating, false otherwise.
        """
        g = ContextFree.remove_nongenerating_nonterminals(grammar, remove)
        return g.start is not None

    @staticmethod
    def remove_unreachable_symbols(grammar, inplace=False):
        # type: (Grammar, bool) -> Grammar
        """
        Remove unreachable symbols from the gramar
        :param grammar: Grammar where to symbols remove
        :param inplace: True if transformation should be performed in place. False by default.
        :return: Grammar without unreachable symbols.
        """
        return remove_unreachable_symbols(grammar, inplace)

    @staticmethod
    def remove_useless_symbols(grammar, inplace=False):
        # type: (Grammar, bool) -> Grammar
        """
        Remove useless symbols from the grammar.
        Useless symbols are unreachable or nongenerating one.
        :param grammar: Grammar where to symbols remove
        :param inplace: True if transformation should be performed in place, false otherwise.
        False by default.
        :return: Grammar without useless symbols.
        """
        grammar = ContextFree.remove_nongenerating_nonterminals(grammar, inplace)
        grammar = ContextFree.remove_unreachable_symbols(grammar, True)
        return grammar

    @staticmethod
    def remove_rules_with_epsilon(grammar, inplace=False):
        # type: (Grammar, bool) -> Grammar
        """
        Remove epsilon rules.
        :param grammar: Grammar where rules remove
        :param inplace: True if transformation should be performed in place, false otherwise.
        False by default.
        :return: Grammar without epsilon rules.
        """
        return remove_rules_with_epsilon(grammar, inplace)

    @staticmethod
    def find_nonterminals_rewritable_to_epsilon(grammar):
        # type: (Grammar) -> Dict[Type[Nonterminal], Type[Rule]]
        """
        Get nonterminals rewritable to epsilon.
        :param grammar: Grammar where to search.
        :return: Dictionary, where key is nonterminal rewritable to epsilon and
        value is rule that is responsible for it.
        The rule doesn't need to rewrite to epsilon directly,
        but the whole right side can be rewritable to epsilon using different rules.
        """
        return find_nonterminals_rewritable_to_epsilon(grammar)

    @staticmethod
    def find_nonterminals_reachable_by_unit_rules(grammar):
        # type: (Grammar) -> UnitSymbolReachability
        """
        Get nonterminal for which exist unit rule
        :param grammar: Grammar where to search
        :return: Instance of UnitSymbolReachability.
        """
        return find_nonterminals_reachable_by_unit_rules(grammar)

    @staticmethod
    def remove_unit_rules(grammar, inplace=False):
        # type: (Grammar, bool) -> Grammar
        """
        Remove unit rules from the grammar.
        :param grammar: Grammar where remove the rules.
        :param inplace: True if transformation should be performed in place. False by default.
        :return: Grammar without unit rules.
        """
        return remove_unit_rules(grammar, inplace)

    @staticmethod
    def transform_to_chomsky_normal_form(grammar, inplace=False):
        # type: (Grammar, bool) -> Grammar
        """
        Transform grammar to Chomsky Normal Form.
        :param grammar: Grammar to transform.
        :param inplace: True if transformation should be performed in place. False by default.
        :return: Grammar in Chomsky Normal Form.
        """
        return transform_to_chomsky_normal_form(grammar, inplace)

    @staticmethod
    def prepare_for_cyk(grammar, inplace=False):
        # type: (Grammar, bool) -> Grammar
        """
        Take common context-free grammar and perform all the necessary steps to use it in the CYK algorithm.
        Performs following steps:
        - remove useless symbols
        - remove rules with epsilon
        - remove unit rules
        - remove useless symbols once more (as previous steps could change the grammar)
        - transform it to Chomsky Normal Form
        :param grammar: Grammar to transform.
        :param inplace: True if the operation should be done in place. False by default.
        :return: Modified grammar.
        """
        grammar = ContextFree.remove_useless_symbols(grammar, inplace)
        grammar = ContextFree.remove_rules_with_epsilon(grammar, True)
        grammar = ContextFree.remove_unit_rules(grammar, True)
        grammar = ContextFree.remove_useless_symbols(grammar, True)
        grammar = ContextFree.transform_to_chomsky_normal_form(grammar, True)
        return grammar



#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 02.08.2017 18:15
:Licence MIT
Part of grammpy

"""

from .RuleSyntaxException import RuleSyntaxException


class NonterminalDoesNotExistsException(RuleSyntaxException):
    """
    Nonterminal does not exists in provided grammar
    """
    def __init__(self, rule, nonterminal, grammar):
        super().__init__(rule, 'Nonterminal does not exists in current grammar')
        self.nonterminal = nonterminal
        self.grammar = grammar



#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 04.08.2017 19:00
:Licence MIT
Part of grammpy

"""
from setuptools import setup

# read the contents of your README file
from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

v = '2.0.0'

setup(
    name='grammpy',
    version=v,
    packages=[
        'grammpy',
        'grammpy.exceptions',
        'grammpy.old_api',
        'grammpy.parsers',
        'grammpy.parsers.CYK',
        'grammpy.representation',
        'grammpy.representation.support',
        'grammpy.transforms',
        'grammpy.transforms.ChomskyForm',
        'grammpy.transforms.EpsilonRulesRemove',
        'grammpy.transforms.NongeneratingSymbolsRemove',
        'grammpy.transforms.SplittedRules',
        'grammpy.transforms.UnitRulesRemove',
        'grammpy.transforms.UnreachableSymbolsRemove',
    ],
    url='https://github.com/PatrikValkovic/grammpy',
    download_url='https://github.com/PatrikValkovic/grammpy/archive/v' + v + '.tar.gz',
    license='The MIT License',
    author='Patrik Valkovic',
    author_email='patrik.valkovic@hotmail.cz',
    description='CYK library with all required tools to parse context-free grammars.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    install_requires=[
        'typing',
        'deprecated',
    ],
)



#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 28.07.2017 09:14
:Licence MIT
Part of grammpy

"""

from .support._RuleConnectable import _RuleConnectable


class Nonterminal(_RuleConnectable):
    '''
    Base class that represents nonterminals.
    '''
    pass



#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 09.03.2019 17:31
:Licence MIT
Part of grammpy

"""
from inspect import isclass
from typing import Type, TYPE_CHECKING

from ... import Nonterminal

if TYPE_CHECKING:  # pragma: no cover
    from ... import Rule


def _is_unit(rule):
    # type: (Type[Rule]) -> bool
    """
    Check if parameter is unit rule.
    :param rule: Object to check.
    :return: True if is parameter unit rule, false otherwise.
    """
    return len(rule.left) == 1 and len(rule.right) == 1 and \
           isclass(rule.fromSymbol) and isclass(rule.toSymbol) and \
           issubclass(rule.fromSymbol, Nonterminal) and issubclass(rule.toSymbol, Nonterminal)



#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 15.10.2018 15:10
:Licence MIT
Part of grammpy

"""
from typing import Iterable, TYPE_CHECKING, Any

from ._BaseSet import _BaseSet

if TYPE_CHECKING:  # pragma: no cover
    from .. import Grammar


class _TerminalSet(_BaseSet):
    """
    Set that store terminals inside the grammar.
    """

    def __init__(self, grammar, assign_map, iterable=None):
        # type: (Grammar, dict, Iterable[Any]) -> None
        """
        Create new instance of _TerminalSet.
        :param grammar: Grammar for which create the set.
        :param assign_map: Map used for assignment rules to terminals.
        :param iterable: Terminals to insert.
        """
        self._grammar = grammar
        self._assign_map = assign_map
        super().__init__()
        self.add(*(iterable or []))

    def add(self, *terminals):
        # type: (Iterable[Any]) -> None
        """
        Add terminals into the set.
        :param terminals: Terminals to insert.
        """
        for term in terminals:
            if term in self:
                continue
            super().add(term)
            self._assign_map[term] = set()

    def remove(self, *terminals):
        # type: (Iterable[Any]) -> None
        """
        Remove terminals from the set.
        Removes also rules using this terminal.
        :param terminals: Terminals to remove.
        :raise KeyError if the object is not in the set.
        """
        for term in set(terminals):
            if term not in self:
                raise KeyError('Terminal ' + str(term) + ' is not inside')
            self._grammar.rules.remove(*self._assign_map[term], _validate=False)
            del self._assign_map[term]
            super().remove(term)


#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 24.01.2019 13:04
:Licence MIT
Part of grammpy

"""
from collections import Iterable
from copy import copy
from typing import List, Union, Type, Optional, TYPE_CHECKING, Any, Generator

from deprecated import deprecated

from ._support import RulesClass
from .. import Grammar as NewGrammar, Terminal, Rule

if TYPE_CHECKING:  # pragma: no cover
    from .. import Nonterminal
    from ..representation.support._NonterminalSet import _NonterminalSet
    from ..representation.support._TerminalSet import _TerminalSet


@deprecated(version='2.0.0', reason='Use new API')
class Grammar:
    """
    Provide base interface for manipulating with the grammar.
    This grammar is deprecated and it's only bridge between the old and new API.
    You should use new API, implementation of this class is less efficient.
    """

    def __init__(self,
                 terminals=None,
                 nonterminals=None,
                 rules=None,
                 start_symbol=None):
        # type: (Optional[Iterable[Any]], Optional[Iterable[Type[Nonterminal]]], Optional[Iterable[Type[Rule]]], Optional[Type[Nonterminal]]) -> Grammar
        """
        Create instance of grammar.
        :param terminals: Sequence of terminals to add, empty sequence by default.
        :param nonterminals: Sequence of nonterminals to add, empty sequence by default.
        :param rules: Sequence of rules to add, empty sequence by default.
        :param start_symbol: Start symbol of the grammar, start symbol must be in nonterminals.
        """
        self._gr = NewGrammar(terminals=terminals,
                              nonterminals=nonterminals,
                              rules=rules,
                              start_symbol=start_symbol)
        self.rules = RulesClass(self)

    # Term part
    def add_term(self, term):
        # type: (Union[Iterable[Any], Any]) -> List[Terminal]
        """
        Add terminal or terminals into grammar.
        :param term: Object or sequence of objects representing terminals.
        :return: List terminals added into grammar as sequence of Terminal instances.
        """
        return list(self._add_term(term))

    def _add_term(self, term):
        # type: (Union[Iterable[Any], Any]) -> Generator[Terminal]
        """
        Add terminal or terminals into grammar.
        :param term: Object or sequence of objects representing terminals.
        :return: List terminals added into grammar as sequence of Terminal instances.
        """
        term = [] if term is None else term
        if not isinstance(term, Iterable) or isinstance(term, str):
            term = [term]
        for t in term:
            if t not in self.terminals:
                self.terminals.add(t)
                yield Terminal(t)

    def remove_term(self, term=None):
        # type: (Optional[Union[Iterable[Any], Any]]) -> List[Terminal]
        """
        Delete terminal or terminals from grammar.
        :param term: Object or sequence of objects representing terminals.
        :return: List of terminals removed from the grammar as sequence of Terminal instances.
        """
        term = self.terminals if term is None else term
        if not isinstance(term, Iterable) or isinstance(term, str):
            term = [term]
        tmp = []
        for t in list(term):
            if t in self.terminals:
                tmp.append(Terminal(t))
            self.terminals.remove(t)
        return tmp

    def have_term(self, term):
        # type: (Union[Iterable[Any], Any]) -> bool
        """
        Check if terminal or terminals are in the grammar.
        :param term: Object or sequence of objects representing terminals.
        :return: True if all objects in the parameter are in the grammar, false otherwise.
        """
        term = [] if term is None else term
        if not isinstance(term, Iterable) or isinstance(term, str):
            term = [term]
        for t in term:
            if t not in self.terminals:
                return False
        return True

    def get_term(self, term=None):
        # type: (Optional[Union[Iterable[Any], Any]]) -> Union[Optional[Terminal], List[Optional[Terminal]]]
        """
        Get terminals stored in grammar that match terminal or terminals passed as parameter.
        :param term: Object or sequence of objects representing terminals.
        :return: List of terminals in the grammar as sequence of Terminal object.
        Terminal or terminals that weren't in the grammar are replaced with None value.
        Return single value if parameter is single value, return list if parameter is iterable.
        If the parameter is None (default), return all terminals.
        """
        if term is None:
            return [Terminal(i) for i in self.terminals]
        is_single = False
        if not isinstance(term, Iterable) or isinstance(term, str):
            term = [term]
            is_single = True
        tmp = []
        for t in term:
            handled = False
            for i in self.terminals:
                if t == i and is_single is not True:
                    handled = True
                    tmp.append(Terminal(i))
                elif t == i:
                    return Terminal(i)
            if not handled and is_single is not True:
                tmp.append(None)
            elif not handled:
                return None
        return tmp

    def term(self, term=None):
        # type: (Optional[Union[Iterable[Any], Any]]) -> Union[Optional[Terminal], List[Optional[Terminal]]]
        """
        Get terminals stored in grammar that match terminal or terminals passed as parameter.
        :param term: Object or sequence of objects representing terminals.
        :return: List of terminals in the grammar as sequence of Terminal object.
        Terminal or terminals that weren't in the grammar are replaced with None value.
        Return single value if parameter is single value, return list if parameter is iterable.
        If the parameter is None (default), return all terminals.
        """
        return self.get_term(term)

    def terms(self):
        # type: () -> List[Terminal]
        """
        Get all terminals within the grammar.
        :return: List of terminals in the grammar as Terminal instances.
        """
        return self.get_term(None)

    def terms_count(self):
        # type: () -> int
        """
        Get number of terminals within the grammar.
        :return: Number of terminals within the grammar.
        """
        return len(self.terminals)

    def terms_clear(self):
        # type: () -> List[Terminal]
        """
        Delete all terminals from grammar.
        :return: List of terminals removed from the grammar as list of Terminal instances.
        """
        return self.remove_term(None)

    # Nonterm part
    def add_nonterm(self, nonterm):
        # type: (Union[Iterable[Type[Nonterminal]], Type[Nonterminal]]) -> List[Type[Nonterminal]]
        """
        Add nonterminal or nonterminals into grammar.
        :param nonterm: Single or sequence of nonterminals.
        :return: List of nonterminals added into the grammar.
        """
        return list(self._add_nonterm(nonterm))

    def _add_nonterm(self, nonterm):
        # type: (Union[Iterable[Type[Nonterminal]], Type[Nonterminal]]) -> Generator[Type[Nonterminal]]
        """
        Add nonterminal or nonterminals into grammar.
        :param nonterm: Single or sequence of nonterminals.
        :return: List of nonterminals added into the grammar.
        """
        nonterm = [] if nonterm is None else nonterm
        if not isinstance(nonterm, Iterable):
            nonterm = [nonterm]
        for t in nonterm:
            if t not in self.nonterminals:
                self.nonterminals.add(t)
                yield t

    def remove_nonterm(self, nonterm=None):
        # type: (Optional[Union[Iterable[Type[Nonterminal]], Type[Nonterminal]]]) -> List[Type[Nonterminal]]
        """
        Delete nonterminal or nonterminals from the grammar.
        :param nonterm: Single or sequence of nonterminals to remove.
        :return: List of nonterminals removed from the grammar.
        """
        nonterm = self.nonterminals if nonterm is None else nonterm
        if not isinstance(nonterm, Iterable):
            nonterm = [nonterm]
        tmp = []
        for t in list(nonterm):
            if t in self.nonterminals:
                tmp.append(t)
                self.nonterminals.remove(t)
        return tmp

    def have_nonterm(self, nonterm):
        # type: (Union[Iterable[Type[Nonterminal]], Type[Nonterminal]]) -> bool
        """
        Check if nonterminal or nonterminals are in the grammar.
        :param nonterm: Single or sequence of nonterminals to check.
        :return: True if all nonterminals in the parameter are in the grammar, false otherwise.
        """
        nonterm = [] if nonterm is None else nonterm
        if not isinstance(nonterm, Iterable):
            nonterm = [nonterm]
        contains = True
        for t in nonterm:
            contains = t in self.nonterminals and contains
        return contains

    def get_nonterm(self, nonterm=None):
        # type: (Optional[Union[Iterable[Type[Nonterminal]], Type[Nonterminal]]]) -> Union[Optional[Type[Nonterminal]], List[Optional[Type[Nonterminal]]]]
        """
        Get nonterminals stored in grammar that match nonterminal or nonterminals passed as parameter.
        :param nonterm: Single or sequence of nonterminals.
        :return: List of nonterminals in the grammar.
        Nonterminal or nonterminals that weren't in the grammar are replaced with None value.
        Return single value if parameter is single value, return list if parameter is iterable.
        If the parameter is None (default), return all nonterminals.
        """
        if nonterm is None:
            return list(self.nonterminals)
        is_single = False
        if not isinstance(nonterm, Iterable):
            nonterm = [nonterm]
            is_single = True
        tmp = []
        for t in nonterm:
            if t in self.nonterminals and is_single is not True:
                tmp.append(t)
            elif t in self.nonterminals:
                return t
            elif is_single is not True:
                tmp.append(None)
            else:
                return None
        return tmp

    def nonterm(self, nonterm=None):
        # type: (Optional[Union[Iterable[Type[Nonterminal]], Type[Nonterminal]]]) -> Union[Optional[Type[Nonterminal]], List[Optional[Type[Nonterminal]]]]
        """
        Get nonterminals stored in grammar that match nonterminal or nonterminals passed as parameter.
        :param nonterm: Single or sequence of nonterminals.
        :return: List of nonterminals in the grammar.
        Nonterminal or nonterminals that weren't in the grammar are replaced with None value.
        Return single value if parameter is single value, return list if parameter is iterable.
        If the parameter is None (default), return all nonterminals.
        """
        return self.get_nonterm(nonterm)

    def nonterms(self):
        # type: () -> List[Type[Nonterminal]]
        """
        Get all nonterminals within the grammar.
        :return: List of nonterminals in the grammar.
        """
        return self.get_nonterm(None)

    def nonterms_count(self):
        # type: () -> int
        """
        Get number of nonterminals in the grammar.
        :return: Number of nonterminals in the grammar.
        """
        return len(self.nonterminals)

    def nonterms_clear(self):
        # type: () -> List[Type[Nonterminal]]
        """
        Delete all nonterminals from grammar.
        :return: List of nonterminals removed from the grammar.
        """
        return self.remove_nonterm(None)

    # Rules part
    def add_rule(self, rules):
        # type: (Union[Iterable[Type[Rule]], Type[Rule]]) -> List[Type[Rule]]
        """
        Add rule or rules into grammar.
        :param rules: Single or sequence of rules.
        :return: List of rules added into the grammar.
        """
        rules = [] if rules is None else rules
        if not isinstance(rules, Iterable):
            rules = [rules]
        return list(self.rules.add(*rules))

    def remove_rule(self, rules=None, *, _validate=True):
        # type: (Optional[Union[Iterable[Type[Rule]], Type[Rule]]], None, bool) -> List[Type[Rule]]
        """
        Delete rule or rules from the grammar.
        :param rules: Single or sequence of rules to remove.
        :param _validate: True if the rule should be validated before deleting.
        This parameter is only for internal use.
        :return: List of rules removed from the grammar.
        """
        rules = self.rules if rules is None else rules
        if not isinstance(rules, Iterable):
            rules = [rules]
        tmp = []
        for t in list(rules):
            if t in self.rules:
                tmp.append(t)
                self.rules.remove(t)
        return tmp

    def have_rule(self, rules):
        # type: (Union[Iterable[Type[Rule]], Type[Rule]]) -> bool
        """
        Check if rule or rules are in the grammar.
        :param rules: Single or sequence of rules to check.
        :return: True if all rules in the parameter are in the grammar, false otherwise.
        """
        rules = [] if rules is None else rules
        if not isinstance(rules, Iterable):
            rules = [rules]
        contains = True
        for t in rules:
            contains = t in self._gr.rules and contains
        return contains

    def get_rule(self, rules=None):
        # type: (Optional[Union[Iterable[Type[Rule]], Type[Rule]]]) -> Union[Optional[Type[Rule]], List[Optional[Type[Rule]]]]
        """
        Get rules stored in grammar that match rule or rules passed as parameter.
        :param rules: Single or sequence of rules.
        :return: List of rules in the grammar.
        Rule or rules that weren't in the grammar are replaced with None value.
        Return single value if parameter is single value, return list if parameter is iterable.
        If the parameter is None (default), return all rules.
        """
        if rules is None:
            return list(self.rules)
        is_single = False
        if not isinstance(rules, Iterable):
            rules = [rules]
            is_single = True
        result = []
        for rule in rules:
            rule.validate(self)
            result = result + self.rules.get(rule)
        return result[0] if is_single and len(result) == 1 else result

    def rule(self, rules=None):
        # type: (Optional[Union[Iterable[Type[Rule]], Type[Rule]]]) -> Union[Optional[Type[Rule]], List[Optional[Type[Rule]]]]
        """
        Get rules stored in grammar that match rule or rules passed as parameter.
        :param rules: Single or sequence of rules.
        :return: List of rules in the grammar.
        Rule or rules that weren't in the grammar are replaced with None value.
        Return single value if parameter is single value, return list if parameter is iterable.
        If the parameter is None (default), return all rules.
        """
        return self.get_rule(rules)

    def _rules(self):
        # type: () -> List[Type[Rule]]
        """
        Get all rules within the grammar.
        :return: List of rules in the grammar.
        """
        return self.get_rule()

    def rules_count(self):
        # type: () -> int
        """
        Get number of rules in the grammar.
        :return: Number of rules in the grammar.
        """
        return len(self._gr.rules)

    def rules_clear(self):
        # type: () -> List[Type[Rule]]
        """
        Delete all rules from grammar.
        :return: List of rules removed from the grammar.
        """
        return self.remove_rule(None)

    # StartSymbol
    def start_get(self):
        # type: () -> Optional[Type[Nonterminal]]
        """
        Get start symbol of the grammar. Default value is None.
        :return: Start symbol of the grammar.
        """
        return self.start

    def start_set(self, nonterminal = None):
        # type: (Optional[Type[Nonterminal]]) -> None
        """
        Set start symbol of the grammar.
        You can set start symbol to None to clear the start symbol.
        :param nonterminal: New start symbol.
        The start symbol needs to be in the nonterminals.
        :raise NonterminalDoesNotExistsException: If the start symbol is not in nonterminals.
        """
        self.start = nonterminal

    def start_isSet(self):
        # type: () -> bool
        """
        Check if the start symbol is set. That mean if the start symbol is not None.
        :return: True if the start symbol is set, false otherwise.
        """
        return self.start is not None

    def start_is(self, nonterminal):
        # type: (Type[Nonterminal]) -> bool
        """
        Check if start symbol is nonterminal passed as parameter.
        :param nonterminal: Nonterminal to check.
        :return: True if start symbol is parameter, false otherwise.
        """
        return self.start == nonterminal

    # Copy
    def __copy__(self):
        # type: () -> Grammar
        """
        Create shallow copy of the grammar.
        :return: Shallow copy of the grammar.
        """
        new_grammar = copy(self._gr)
        i = Grammar()
        i._gr = new_grammar
        return i

    # New API
    @property
    def terminals(self):
        # type: () -> _TerminalSet
        """
        Get set representing terminals.
        :return: Set representing terminals.
        """
        return self._gr.terminals

    @property
    def nonterminals(self):
        # type: () -> _NonterminalSet
        """
        Get set representing nonterminals.
        :return: Set representing nonterminals.
        """
        return self._gr.nonterminals

    @property
    def start(self):
        # type: () -> Optional[Type[Nonterminal]]
        """
        Get start symbol of the grammar.
        :return: Start symbol of the grammar, or None if it isn't defined.
        """
        return self._gr.start

    @start.setter
    def start(self, s):
        # type: (Optional[Type[Nonterminal]]) -> None
        """
        Set start symbol of the grammar.
        :param s: Start symbol to set.
        :raise NonterminalDoesNotExistsException: If the start symbol is not in nonterminals.
        """
        self._gr.start = s

    @start.deleter
    def start(self):
        # type: () -> None
        """
        Unset start symbol and set it to None.
        """
        del self._gr.start



#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 27.04.2018 17:49
:Licence MIT
Part of grammpy

"""

from .ParsingException import ParsingException

class NotParsedException(ParsingException):
    """
    Represent exception that the tree was not parsed successfully
    """
    pass
